
// for Linux/svgalib.
// We set up the screen and mouse here and then we start

#define MAX_MOUSE_CURSIZE 50

#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <ctype.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <semaphore.h>
#include <pthread.h>
#include <sys/time.h>
#include "globdef.h"
#include "vernr.h"
#include "mousedef.h"
#include "fontdef.h"
#include "keyboard_def.h"
#include "uilim.h"



extern int netfd_fft2in;


char mouse_thread_flag;
char *behind_mouse;
int mouse_cursize;
int mouse_xmax;
int mouse_xmin;
int mouse_ymax;
int mouse_ymin;
int mouse_hide_flag;


pthread_t thread_identifier_kill_all;
pthread_t thread_identifier_keyboard;
pthread_t thread_identifier_mouse;
pthread_t thread_identifier_zoom;
pthread_t thread_identifier_fft2;


sem_t sem_kill_all;
sem_t sem_mouse;
sem_t sem_keyboard;

void graphics_init(void);
void thread_zoom(void);
void zoom(void);
int lin_global_uiparms(int wn);
void thread_keyboard(void);
void thread_mouse(void);
void thread_kill_all(void);
void restore_behind_mouse(void);
void ui_setup(void);


int main(int argc, char *argv[])
{
char chr;
int i;
// Make silly use of parameters to suppress warning from compiler
chr=argv[0][0];
i=argc;
// Init svgalib and set it to text mode.
vga_init();
vga_setmode(TEXT);
keyboard_buffer=malloc(KEYBOARD_BUFFER_SIZE*sizeof(int)+
                                4*(MAX_MOUSE_CURSIZE+1)*sizeof(char));
behind_mouse = (void*)keyboard_buffer+KEYBOARD_BUFFER_SIZE*sizeof(int);
mouse_hide_flag=0;
sem_init(&sem_keyboard,0,0);
pthread_create(&thread_identifier_keyboard,NULL,(void*)thread_keyboard, NULL);
pthread_create(&thread_identifier_mouse,NULL,(void*)thread_mouse, NULL);
// Create a thread that will close the entire process in a controlled way
// in case fatal_error() is called or the ESC key is pressed.
sem_init(&sem_kill_all,0,0);
kill_all_flag=FALSE;
error_exit_msg=NULL;
pthread_create(&thread_identifier_kill_all,NULL,(void*)thread_kill_all, NULL);
if(DUMPFILE)
  {
  dmp = fopen("dmp", "w");
  DEB"\n******************************\n");
  }
else
  {  
  dmp=NULL;
  }
os_flag=OS_FLAG_LINUX;
ui_setup();
graphics_init();
if(error_exit_msg != NULL)goto text_exitmain;
vga_setmousesupport(1);
// Set up mouse and system parameters in case it has not been done already.
mouse_cursize=HEIGHT/80;
if(mouse_cursize > MAX_MOUSE_CURSIZE)mouse_cursize=MAX_MOUSE_CURSIZE;
mouse_setxrange(0, WIDTH-1);
mouse_setyrange(0, HEIGHT-1);
mouse_setwrap(MOUSE_NOWRAP);
mouse_x=screen_width/2;
mouse_y=screen_height/2;
new_mouse_x=screen_width/2;
new_mouse_y=screen_height/2;
mouse_setposition( mouse_x, mouse_y);
pthread_create(&thread_identifier_zoom,NULL, (void*)thread_zoom, NULL);
usleep(50000);
pthread_join(thread_identifier_zoom,0);
clear_keyboard();
text_exitmain:;
vga_setmode(TEXT);
printf("\n");
if(error_exit_msg != NULL)printf("\n%s",error_exit_msg);
printf("\nLeaving %s\n",PROGRAM_NAME);
pthread_cancel(thread_identifier_mouse);
pthread_join(thread_identifier_mouse,0);
usleep(20000);
pthread_cancel(thread_identifier_keyboard);
pthread_join(thread_identifier_keyboard,0);
if(dmp!=NULL)fclose(dmp);
return 0;
}

void graphics_init(void)
{
int i;
vga_setmode(ui.vga_mode);
gl_setcontextvga(ui.vga_mode);
if(BYTESPERPIXEL != 1)
  {
  fatal_error("vga mode not 256 colours");
  return;
  }
screen_width=WIDTH;
screen_height=HEIGHT;
if(screen_width < 640 || screen_height < 480)
  {
  fatal_error("Screen too small");
  return;
  }
for(i=0; i<256;i++)
  {
  gl_setpalettecolor(i, svga_palette[3*i+2],
                        svga_palette[3*i+1],
                        svga_palette[3*i  ]);
  }
init_font(ui.font_scale);
}


int force_getchar(void)
{
// Get a character from keyboard, but set a 0.1 s timeout
// and return -1 if no key was pressed.
struct termios term, oterm;
int c = 0;
// get the terminal settings 
tcgetattr(0, &oterm);
// get a copy of the settings, which we modify 
memcpy(&term, &oterm, sizeof(term));
// put the terminal in non-canonical mode, any 
// reads timeout after 0.1 seconds or when a 
// single character is read
term.c_lflag = term.c_lflag & (!ICANON);
term.c_cc[VMIN] = 0;
term.c_cc[VTIME] = 1;
tcsetattr(0, TCSANOW, &term);
// get input - timeout after 0.1 seconds or
// when one character is read. If timed out
// getchar() returns -1, otherwise it returns
// the character
c=getchar();
//   reset the terminal to original state 
tcsetattr(0, TCSANOW, &oterm);
return c;
}


void thread_keyboard(void)
{
int c, i, k;
char s[80];
pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,&i);
while(1)
  {
// It is silly to use force_getchar here. 
// We should set up stdin so it will return on each char of an ESC sequence.
// so we would not have to loop around at 10 Hz here.
  c=getchar();
  keyboard_buffer[keyboard_buffer_ptr]=c;
  if(c == 27)
    {
// Set the flag and wait a while to make it unlikely
// it will not be visible for the kill_all thread.
// We must not kill the keyboard thread while force_getchar() is
// executed because the changed terminal setting will survive
// after Linrad is finished and make the command line behave
// incorrectly.
      {
      c=force_getchar();
      if(c == -1)
        {
// The ESC key was pressed. Exit from Linrad NOW!    
        sem_post(&sem_kill_all);
        while(!kill_all_flag)usleep(3000);
        keyboard_buffer[keyboard_buffer_ptr]=0;
        keyboard_buffer_ptr=(keyboard_buffer_ptr+1)&
                                                  (KEYBOARD_BUFFER_SIZE-1);
        sem_post(&sem_keyboard);
        goto kbd_exit;
        }    
// The ESC sequences are all [27],[91],.....
      if(c == 91)
        {
        c=force_getchar();
// Arrow keys are:
// ARROW_UP=65
// ARROW_DOWN=66
// ARROW_RIGHT=67
// ARROW_LEFT=68
        if(c >= 65 && c <= 68)
          {
          c+=ARROW_UP_KEY-65;
          goto fkn_key_ok;
          }
        if(c >= 49 && c <= 54)
          {
          i=force_getchar();
// These keys have to be followed by 126
// HOME=49
// INSERT=50
// DELETE=51
// END=52
// PAGE_UP=53
// PAGE_DOWN=54
          if(i == 126)
            {
            c+=HOME_KEY-49;
            goto fkn_key_ok;
            }
          k=force_getchar();
          if(k == 126)
            {  
            if(c == 49)
              {
// F6=55
// F7=56
// F8=57
              if(i >= 55 && i <= 57)
                {
                c=i+F6_KEY-55;
                goto fkn_key_ok;
                }
              }  
            if(c == 50)
              {
// F9=48
// F10=49
// F10_PAD=50
// F11=51
// F12=52
// SHIFT_F1=53
// SHIFT_F2=54
// SHIFT_F2_PAD=55
// SHIFT_F3=56
// SHIFT_F4=57
              if(i >= 48 && i <= 57)
                { 
                c=i+F9_KEY-48;
                goto fkn_key_ok;
                }
              }    
            if(c == 51)
              {
// SHIFT_F5=49
// SHIFT_F6=50
// SHIFT_F7=51
// SHIFT_F8=52
             if(i >= 49 && i <= 52)
                { 
                c=i+SHIFT_F5_KEY-49;
                goto fkn_key_ok;
                }
              }
            }    
          sprintf(s,"ESC sequence):[27][91][%d][%d][%d]",c,i,k);
          goto skip;
          }
// PAUSE = 80
        if(c == 80)
          {
          c=PAUSE_KEY;
          goto fkn_key_ok;
          }
        if(c == 91)
          {

          c=force_getchar();
// F1=65
// F2=66
// F3=67
// F4=68
// F5=69
          if(c >= 65 && c <= 69)
            {
            c+=F1_KEY-65;
            goto fkn_key_ok;
            }
          sprintf(s,"ESC sequence:[27][91][91][%d]",c);
          goto skip;
          }
        sprintf(s,"ESC sequence:[27][91][%d]",c);
        goto skip;
        }
// ESC followed by something else than 91 (should never occur)
      sprintf(s,"ESC sequence:[27][%d]",c);
skip:;
      lir_text(1,5,s);
      c=force_getchar();
      if(c != -1)
        {
        i=0;
        while(s[i]!=0 && i<70)i++;
        sprintf(&s[i],"[%d]",c);
        goto skip;
        }
      if(c == -1) goto next;
fkn_key_ok:;
      keyboard_buffer[keyboard_buffer_ptr]=c;
      }
    }
  keyboard_buffer_ptr=(keyboard_buffer_ptr+1)&(KEYBOARD_BUFFER_SIZE-1);
  sem_post(&sem_keyboard);
next:;
  }
kbd_exit:;
}

void thread_mouse(void)
{
int i, button;
pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,&i);
while(TRUE)
  {
// **********************************************
// Wait for svgalib to report a mouse event
  mouse_waitforupdate();
// We do nothing unless mouse_flag is set.
  if(mouse_flag != 0)
    {
// Use svgalib functions to get mouse data
// First wait for a mouse event.
    new_mouse_x= mouse_getx();
    new_mouse_y= mouse_gety();
    button= mouse_getbutton();
    if (button & MOUSE_LEFTBUTTON) 
      {
      new_lbutton_state=1;
      }
    else
      {
      new_lbutton_state=0;
      }
    if (button & MOUSE_RIGHTBUTTON) 
      {
      new_rbutton_state=1;
      }
    else
      {
      new_rbutton_state=0;
      }
    check_mouse_actions();
    }
  }
}

void thread_zoom(void)
{
netfd_fft2in=-1;
zoom();
if(netfd_fft2in != -1)close(netfd_fft2in);
}

int lin_global_uiparms(int wn)
{
char ss[10];
char *errms;
int line;
char s[80];
int *modflag;
int i, grp;
vga_modeinfo *info;
line=0;
// ************************************************
// *****   Select screen number for svgalib  ******
// ************************************************
modflag = malloc((GLASTMODE+1) * sizeof(int));
if(modflag == NULL)
  {
  if(wn==0)
    {
    exit(0);
    }
  else
    {  
    fatal_error(out_of_memory);
    return FALSE;
    }
  }
modflag[0]=0;
sprintf(s,"Choose one of the following video modes:");
if(wn==0)
  {
  for(i=0; i<32; i++)printf("\n");
  printf("%s",s);
  }
else
  {
  clear_screen();
  settextcolor(15);
  lir_text(0,line,s);
  settextcolor(7);
  line++;
  }  
grp=0;
for (i = 1; i <= GLASTMODE; i++)
  {
  if (vga_hasmode(i))
    {
    info = vga_getmodeinfo(i);
    if (info->colors == 256)
      {
      modflag[0]++;
      modflag[i]=1;
      sprintf(s,"%5d: %dx%d       ",i, info->width, info->height);
      s[20]=0;
      if(wn==0)
        { 
        if(grp==0)printf("\n");
        printf("%s",s); 
        }
      else  
        {
        lir_text(20*grp,line,s);
        }
      grp++;
      if(grp >= 3)
        {
        grp=0;
        line++;
        if(line > screen_last_line-1)line=screen_last_line-1;
        }
      }
    else
      {
      modflag[i]=0;
      }  
    }
  }
if(wn==0)printf("\n");
line++;
if(modflag[0] == 0)
  {
  sprintf(s,"No 256 color graphics found.");
  if(wn==0)
    {
    printf("%s\n",s);
    exit(0);
    }
  else
    {
    fatal_error(s);
    return FALSE;
    }
  }        
gtmode:;
sprintf(s,"Enter mode number.");
if(wn==0)
  {
  printf("%s Then %s\n",s,press_enter);
  fflush(stdout);
  usleep(100000);
  clear_keyboard();
  i=0;
  ss[0]=' ';
  ss[1]=0;
  while(ss[i]!=10)
    {
    if(i<10)i++;
    await_keyboard();
    if(lir_inkey==0)exit(0);
    ss[i]=lir_inkey;
    ss[i+1]=0;
    }
  sscanf(ss,"%d", &ui.vga_mode);
  }  
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  lir_text(22,line,"=>");
  ui.vga_mode=lir_get_integer(15,line,3,0,GLASTMODE);
  }
sprintf(s,"Error: Mode number out of range");
if(ui.vga_mode <0 || ui.vga_mode > GLASTMODE)goto moderr;
if (modflag[ui.vga_mode] == 0)
  {
moderr:;  
  if(wn==0)
    {
    printf("%s\n\n",s);
    }
  goto gtmode;
  }     
free(modflag);
if(wn!=0)
  {
  line=2;
  clear_screen();
  }
// ********************* FONT SIZE **********************
sprintf(s,"Enter font scale (1 to 5)"); 
fntc:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  await_keyboard();
  if(lir_inkey==0)exit(0);
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  await_keyboard();
  if(kill_all_flag) return FALSE;
  if(line>=screen_last_line)line=screen_last_line;
  line++;
  }
if(lir_inkey < '1' || lir_inkey > '5')goto fntc;
ui.font_scale=lir_inkey-'0';

// *********************** BASE PORT ********************************
sprintf(s,"Enter network base port (%d to %d) =>",
                                  NET_MULTI_MINPORT, NET_MULTI_MAXPORT);
basepo:;
if(wn==0)
  {
  printf("%s Then %s\n",s,press_enter);
  fflush(stdout);
  usleep(100000);
  if(lir_inkey==0)exit(0);
  clear_keyboard();
  i=0;
  ss[0]=' ';
  ss[1]=0;
  while(ss[i]!=10)
    {
    if(i<10)i++;
    await_keyboard();
    if(lir_inkey==0)exit(0);
    ss[i]=lir_inkey;
    ss[i+1]=0;
    }
  sscanf(ss,"%d", &ui.base_port);
  }  
else
  {
  lir_text(0,line,s);
  lir_text(55,line,"=>");
  ui.base_port=lir_get_integer(58,line,5, 
                                    NET_MULTI_MINPORT, NET_MULTI_MAXPORT);
  line++;
  }
ui.base_port=(ui.base_port/10)*10;
if(ui.base_port < NET_MULTI_MINPORT  || 
   ui.base_port > NET_MULTI_MAXPORT)goto basepo;
// **************************** GROUP LAST BYTE **************************
sprintf(s,"[%s] Enter new last byte (0 to 15)", netrec_rx_multi_group);
addr:;
if(wn==0)
  {
  printf("%s Then %s\n",s,press_enter);
  fflush(stdout);
  usleep(100000);
  if(lir_inkey==0)exit(0);
  clear_keyboard();
  i=0;
  ss[0]=' ';
  ss[1]=0;
  while(ss[i]!=10)
    {
    if(i<10)i++;
    await_keyboard();
    if(lir_inkey==0)exit(0);
    ss[i]=lir_inkey;
    ss[i+1]=0;
    }
  sscanf(ss,"%d", &ui.netaddr_offset);
  }  
else
  {
  lir_text(0,line,s);
  lir_text(55,line,"=>");
  ui.netaddr_offset=lir_get_integer(63,line,2,0,15);
  if(kill_all_flag) return FALSE;
  line++;
  }
if(ui.netaddr_offset < 0 || ui.netaddr_offset > 15)goto addr;  
// ********************** SCRATCH FILE SIZE ********************
sprintf(s,"Enter max scratch file size in powers of two(%d to %d)",
                                                     MIN_FILESIZ,MAX_FILESIZ);
filemax:;
if(wn==0)
  {
  printf("%s Then %s\n",s,press_enter);
  fflush(stdout);
  usleep(100000);
  if(lir_inkey==0)exit(0);
  clear_keyboard();
  i=0;
  ss[0]=' ';
  ss[1]=0;
  while(ss[i]!=10)
    {
    if(i<10)i++;
    await_keyboard();
    if(lir_inkey==0)exit(0);
    ss[i]=lir_inkey;
    ss[i+1]=0;
    }
  sscanf(ss,"%d", &ui.max_filesiz);
  if(ui.max_filesiz < MIN_FILESIZ || ui.max_filesiz > MAX_FILESIZ)goto filemax;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.max_filesiz=lir_get_integer(49,line,3,MIN_FILESIZ,MAX_FILESIZ);
  if(kill_all_flag) return FALSE;
  line++;  
  }
// ********************** WATERFALL LINES LIMITATION *************
sprintf(s,"Enter waterfall lines limit (50 and above)");
linemax:;
if(wn==0)
  {
  printf("%s Then %s\n",s,press_enter);
  fflush(stdout);
  usleep(100000);
  if(lir_inkey==0)exit(0);
  clear_keyboard();
  i=0;
  ss[0]=' ';
  ss[1]=0;
  while(ss[i]!=10)
    {
    if(i<10)i++;
    await_keyboard();
    if(lir_inkey==0)exit(0);
    ss[i]=lir_inkey;
    ss[i+1]=0;
    }
  sscanf(ss,"%d", &ui.max_lines);
  if(ui.max_lines < 50)goto linemax;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.max_filesiz=lir_get_integer(49,line,4,50,10000);
  if(kill_all_flag) return FALSE;
  }
// ******************************************************
if(ui.screen_width_factor < MIN_SCREEN_W ||
   ui.screen_width_factor > MAX_SCREEN_W)
  {
  ui.screen_width_factor = (2*MIN_SCREEN_W+MAX_SCREEN_W)/3;
  }
if(ui.screen_height_factor < MIN_SCREEN_H ||
   ui.screen_height_factor > MAX_SCREEN_H)
  {
  ui.screen_height_factor = (2*MIN_SCREEN_H+MAX_SCREEN_H)/3;
  }
errms=save_uiparm();
if(errms != NULL)
  {
  if(wn==0)
    {
    printf("\n%s",errms);
    }
  else
    {
    lir_text(0,8,errms);
    lir_text(15,9,press_any_key);
    await_keyboard();
    }
  return FALSE;
  }
if(ui.vga_mode == 0)ui.vga_mode=10;
if(ui.screen_height_factor ==0)ui.screen_height_factor=40;      
return TRUE;
}

void aw_keyb(void)
{
while(keyboard_buffer_ptr == keyboard_buffer_used)
  {
  if(keyboard_buffer_ptr == keyboard_buffer_used)
    {
    sem_wait(&sem_keyboard);
    if(kill_all_flag) return;
    }
  }
lir_inkey = keyboard_buffer[keyboard_buffer_used];
keyboard_buffer_used=(keyboard_buffer_used+1)&(KEYBOARD_BUFFER_SIZE-1);
}


void await_keyboard(void)
{
lir_refresh_screen();
aw_keyb();
}

void fatal_error(char *errmsg)
{
char *errms;
if(kill_all_flag) return;
errms=(void*)font8x12;
sprintf(errms,"FATAL ERROR:[%s]",errmsg);
DEB"\n%s",errms);
if(dmp != 0)fflush(dmp);
error_exit_msg=errms;
sem_post(&sem_kill_all);
usleep(100000);
}

void thread_kill_all(void)
{
// Wait until the semaphore is released.
// Then stop all processing threads so main can write any
// error message and exit.
sem_wait(&sem_kill_all);
kill_all_flag=TRUE;
}

void test_keyboard(void)
{
if(kill_all_flag!=0)return;
if(keyboard_buffer_ptr != keyboard_buffer_used)
  {
  sem_wait(&sem_keyboard);
  lir_inkey = keyboard_buffer[keyboard_buffer_used];
  keyboard_buffer_used=(keyboard_buffer_used+1)&(KEYBOARD_BUFFER_SIZE-1);
  }
else
  {
  lir_inkey=0;  
  }
}

void clear_screen(void)
{
vga_clear();
}

void lir_refresh_screen(void)
{
// This is used by Windows and X11 to force a copy from memory
// to the screen. svgalib is fast enough to allow
// direct screen updates. (SetPixel under Windows is extremely slow)
}

void lir_fillbox(int x, int y,int w, int h, unsigned char c)
{
gl_fillbox(x,y,w,h,c);
}

void lir_move_mouse_cursor(void)
{
if(mouse_hide_flag !=0)restore_behind_mouse();
set_button_coordinates();
mouse_xmax=mouse_x+mouse_cursize;
mouse_xmin=mouse_x-mouse_cursize;
mouse_ymax=mouse_y+mouse_cursize;
mouse_ymin=mouse_y-mouse_cursize;
}

void restore_behind_mouse(void)
{
int i,imax;
imax=screen_width-mouse_x;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  lir_setpixel(mouse_x+i,mouse_y,behind_mouse[4*i  ]);
  }
imax=mouse_x;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  lir_setpixel(mouse_x-i,mouse_y,behind_mouse[4*i+1]);
  }
imax=screen_height-mouse_y;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  lir_setpixel(mouse_x,mouse_y+i,behind_mouse[4*i+2]);
  }
imax=mouse_y;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  lir_setpixel(mouse_x,mouse_y-i,behind_mouse[4*i+3]);
  }
lir_setpixel(mouse_x,mouse_y,behind_mouse[0]);
mouse_hide_flag=0;
}


void hide_mouse(int x1,int x2,int iy1,int y2)
{
if(mouse_hide_flag ==0)return;
if(mouse_xmax < x1)return;
if(mouse_xmin > x2)return;
if(mouse_ymax < iy1)return;
if(mouse_ymin > y2)return;
restore_behind_mouse();
}

void show_mouse(void)
{
int i,imax;
if(mouse_hide_flag == 1)return;
mouse_hide_flag=1;
imax=screen_width-mouse_x;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  behind_mouse[4*i  ]=gl_getpixel(mouse_x+i,mouse_y);
  lir_setpixel(mouse_x+i,mouse_y,15);
  }
imax=mouse_x;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  behind_mouse[4*i+1]=gl_getpixel(mouse_x-i,mouse_y);
  lir_setpixel(mouse_x-i,mouse_y,15);
  }
imax=screen_height-mouse_y;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  behind_mouse[4*i+2]=gl_getpixel(mouse_x,mouse_y+i);
  lir_setpixel(mouse_x,mouse_y+i,15);
  }
imax=mouse_y;
if(imax > mouse_cursize) imax=mouse_cursize;
for(i=1; i<imax; i++)
  {
  behind_mouse[4*i+3]=gl_getpixel(mouse_x,mouse_y-i);
  lir_setpixel(mouse_x,mouse_y-i,15);
  }
behind_mouse[0]=gl_getpixel(mouse_x,mouse_y);
lir_setpixel(mouse_x,mouse_y,15);
}

void lir_setpixel(int x, int y, unsigned char c)
{
gl_setpixel( x, y, (int)c);
}

void lir_putbox(int x, int y, int w, int h, void* dp)
{
gl_putbox(x,y,w,h,dp);
}

void lir_hline(int x1, int y, int x2, unsigned char c)
{
gl_hline( x1, y, x2, c);
}

void lir_line(int x1, int yy1, int x2,int y2, unsigned char c)
{
gl_line( x1, yy1, x2, y2, c);
}

void lir_sched_yield(void)
{
sched_yield();
}

void lir_sleep(int us)
{
usleep(us);
}

void ui_setup(void)
{
int i, j, k, nread;
char *parinfo;
FILE *file;
int *uiparm;
uiparm=(int*)(&ui);
file = fopen(userint_filename, "rb");
if (file == NULL)
  {
  printf("\nSetup file %s missing.",userint_filename);
full_setup:;
  for(i=0; i<MAX_UIPARM; i++) uiparm[i]=0;
  printf("\nForcing setup to create a new %s file.",userint_filename);
  printf("\n\nPress S to continue");
  printf("\nAny other key to exit."); 
  printf("\nThen press enter\n\n=>"); 
  toupper_await_keyboard();
  if(lir_inkey != 'S') exit(0);
  if(lin_global_uiparms(0) == FALSE)exit(0);
  }
else
  {
  parinfo=malloc(4096);
  if(parinfo == NULL)
    {
    printf("\n\n%s",out_of_memory);
    exit(0);
    }
  for(i=0; i<4096; i++) parinfo[i]=0;
  nread=fread(parinfo,1,4095,file);
  fclose(file);
  file=NULL;
  if(nread >= 4095)
    {
    goto full_setup;
    }
  k=0;
  for(i=0; i<MAX_UIPARM; i++)
    {
    while(parinfo[k]==' ' ||
          parinfo[k]== '\n' )
      {
      k++;
      if(k>=nread-1)goto full_setup;
      }
    j=0;
    while(parinfo[k]== uiparm_text[i][j] && k<nread)
      {
      k++;
      j++;
      } 
    if(uiparm_text[i][j] != 0)goto full_setup;
    while(parinfo[k]!='[')k++;
    sscanf(&parinfo[k],"[%d]",&uiparm[i]);
    while(parinfo[k]!='\n')k++;
    }
  free(parinfo);  
  if (vga_hasmode(ui.vga_mode)==0)goto full_setup;
  if(ui.max_filesiz < MIN_FILESIZ || ui.max_filesiz > MAX_FILESIZ)
    {
    printf("\nIllegal value for max scratch file size");
    goto full_setup;
    }
  if(ui.base_port < NET_MULTI_MINPORT || ui.base_port > NET_MULTI_MAXPORT)
    {
    printf("\nBase port out of range");    
    goto full_setup;
    }
  if(ui.base_port!=(ui.base_port/10)*10)
    {
    printf("\nIllegal base port. Must be a multiple of 10.");
    goto full_setup;
    }
  if(ui.netaddr_offset <0 || ui.netaddr_offset > 15)
    {
    printf("\nNet address offset out of range");
    goto full_setup;
    }
  }
}

void create_netrec_fft2_thread(void)
{
pthread_create(&thread_identifier_fft2,NULL,
                              (void*)thread_rx_fft2_netinput, NULL);
}

void wait_for_netrec_exit(void)
{
pthread_join(thread_identifier_fft2,0);
}

void xz(char *s)
{
if(dmp == NULL) return;
PERMDEB" %s\n",s);
fflush(dmp);
sync();
}
