
#include <ctype.h>
#include <unistd.h>
#include "globdef.h"

#if(OSNUM == OS_FLAG_WINDOWS)
#include <windows.h>
#include <winsock.h>
unsigned int netfd_fft2in;
#define RECV_FLAG 0
typedef struct{
struct in_addr imr_multiaddr;   /* IP multicast address of group */
struct in_addr imr_interface;   /* local IP address of interface */
}IP_MREQ;
WSADATA wsadata;
#define IP_ADD_MEMBERSHIP 12
#define INVSOCK INVALID_SOCKET
#endif

#if(OSNUM == OS_FLAG_LINUX)
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netdb.h>
#include <semaphore.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
int netfd_fft2in;
#define RECV_FLAG MSG_WAITALL
#define IP_MREQ struct ip_mreq
#define INVSOCK -1
#endif


#include "netdef.h"
#include "fft2def.h"
#include "fontdef.h"

#define FFT2_PORT_OFFSET 5
struct sockaddr_in netrec_addr;
struct sockaddr_in netmaster_addr;
fd_set rx_input_fds;
int net_no_of_errors;

void net_input_error(void)
{
char s[80];
net_no_of_errors++;
if(net_no_of_errors > 2)
  {
  sprintf(s,"Network error %d",net_no_of_errors-2);
  lir_text(15,screen_last_line,s);
  }
}
        
void lir_netread_rxin(void *buf)
{
NET_RX_STRUCT *msg; 
char s[80];
fd_set testfds;
struct timeval timeout;
int nbytes;
msg=(void*)buf;
while( !kill_all_flag )
  {
  testfds=rx_input_fds;
  timeout.tv_sec=0;
  timeout.tv_usec=300000;
  if( select(FD_SETSIZE, &testfds, (fd_set *)0, 
                                         (fd_set *)0, &timeout) > 0)
    {   
    if(FD_ISSET(netfd_fft2in,&testfds))
      {
      nbytes=recv(netfd_fft2in,buf,sizeof(NET_RX_STRUCT), RECV_FLAG);
      if(nbytes != sizeof(NET_RX_STRUCT))
        {
        sprintf(s,"NET RX ERROR. Received %d  Expected %d",
                                      nbytes, sizeof(NET_RX_STRUCT));
        lir_text(0,10,s);
        lir_text(0,11,"Incompatible Linrad version?");
        }
      if(passband_center != msg->passband_center)
        {
        passband_center = msg->passband_center;
        }        
      return;
      }  
    }
  }  
}

void thread_rx_fft2_netinput(void)
{
int i;
float *net_floatbuf;
short int *net_shibuf;
struct sockaddr_in rx_addr;
unsigned int addrlen;
int initial_skip_flag;
unsigned int netrd_fft2_ia;
int netrd_fft2_ib;
unsigned short int block_cnt;
net_floatbuf=(void*)&net_rxdata_fft2.buf[0];
net_shibuf=(void*)&net_rxdata_fft2.buf[0];
initial_skip_flag=1;
addrlen=sizeof(rx_addr);
netrd_fft2_ia=0;        
lir_netread_rxin(&net_rxdata_fft2);
if(kill_all_flag) goto fft2net_rxin_error_exit;
netrd_fft2_ib=net_rxdata_fft2.ptr;
block_cnt=net_rxdata_fft2.block_no;
while(!kill_all_flag)
  {
  if(swfloat)
    {
    while(sizeof(float)*netrd_fft2_ia<NET_MULTICAST_PAYLOAD)
      {
      fft2_float[fft2_pa+netrd_fft2_ib  ]=net_floatbuf[netrd_fft2_ia  ];
      netrd_fft2_ib++;
      netrd_fft2_ia++;
      if(netrd_fft2_ib >= fft2_block)
        {
        netrd_fft2_ib=0;
        fft2_pa=(fft2_pa+fft2_block)&fft2_mask;  
        lir_sched_yield();
        }
      }
    }
  else
    {
    while(sizeof(short int)*netrd_fft2_ia<NET_MULTICAST_PAYLOAD)
      {
      fft2_shi[fft2_pa+netrd_fft2_ib  ]=net_shibuf[netrd_fft2_ia  ];
      netrd_fft2_ib++;
      netrd_fft2_ia++;
      if(netrd_fft2_ib >= fft2_block)
        {
        netrd_fft2_ib=0;
        fft2_pa=(fft2_pa+fft2_block)&fft2_mask;  
        lir_sched_yield();
        }
      }
    }      
  netrd_fft2_ia=0;        
  lir_netread_rxin(&net_rxdata_fft2);
  if(kill_all_flag) goto fft2net_rxin_error_exit;
  block_cnt++;
  if( block_cnt != net_rxdata_fft2.block_no || 
                                     netrd_fft2_ib!=net_rxdata_fft2.ptr)
    {
// The block number (or data pointer) is incorrect. Some data is lost!!
    net_input_error();
    i=net_rxdata_fft2.block_no-block_cnt;
    i=(i+0x10000)&0xffff;
    i*=NET_MULTICAST_PAYLOAD/sizeof(float);
    i+=net_rxdata_fft2.ptr-netrd_fft2_ib;
    if(i>0 && i<2*fft2_block)
      {
// We have lost less than two transforms.
// Fill in zeroes and update pointers. The user may not notice.....      
      if(swfloat)
        {
        while(i>0)
          {
          i--;
          fft2_float[fft2_pa+netrd_fft2_ib  ]=0;
          netrd_fft2_ib++;
          netrd_fft2_ia++;
          if(netrd_fft2_ia>=NET_MULTICAST_PAYLOAD/sizeof(float))
            {
            netrd_fft2_ia=0;
            block_cnt++;
            }
          if(netrd_fft2_ib >= fft2_block)
            {
            netrd_fft2_ib=0;
            fft2_pa=(fft2_pa+fft2_block)&fft2_mask;  
            lir_sched_yield();
            }
          }
        }  
      else  
        {
        while(i>0)
          {
          i--;
          fft2_shi[fft2_pa+netrd_fft2_ib  ]=0;
          netrd_fft2_ib++;
          netrd_fft2_ia++;
          if(netrd_fft2_ia>=NET_MULTICAST_PAYLOAD/sizeof(float))
            {
            netrd_fft2_ia=0;
            block_cnt++;
            }
          if(netrd_fft2_ib >= fft2_block)
            {
            netrd_fft2_ib=0;
            fft2_pa=(fft2_pa+fft2_block)&fft2_mask;  
            lir_sched_yield();
            }
          }
        }
      }
    else
      {  
      block_cnt=net_rxdata_fft2.block_no;
      netrd_fft2_ib=net_rxdata_fft2.ptr;
      }
    }
  }
fft2net_rxin_error_exit:;
}

int open_network(void)
{
int setupflag;
int setupcount;
fd_set fds, testfds;
char s[80];
int i, k;
unsigned int addrlen;
unsigned short int portno;
IP_MREQ mreq;
unsigned char *ip;
char *errms;
int nbytes;
struct timeval timeout;
char msgx[sizeof(NET_RX_STRUCT)+2];
NET_RX_STRUCT *msg;
msg=(void*)msgx;
errms=(void*)font8x13;
if(ui.netaddr_offset < 0 || ui.netaddr_offset > 15)
  {
  fatal_error("Network addr out of range");
  return FALSE;
  }
sprintf(&netrec_rx_multi_group[10],"%d",ui.netaddr_offset);
#if(OSNUM == OS_FLAG_WINDOWS)
if(WSAStartup(2, &wsadata) != 0)
  {
  i=WSAGetLastError();
  sprintf(errms,"WSAStartup(...) failed with code %d",i);
  return FALSE;   
  }
#endif
if((netfd_fft2in=socket(AF_INET,SOCK_DGRAM,0)) == INVSOCK) 
  {
  fatal_error("Failed to create socket");
  return FALSE;
  }
// Set a big buffer. Defaults: Linux=109568 and Windows=8192.
i=131072;
k=sizeof(int);
setsockopt(netfd_fft2in,SOL_SOCKET,SO_RCVBUF,(char*)&i,k);
// allow multiple sockets to use the same PORT number 
k=1;
if(setsockopt(netfd_fft2in,SOL_SOCKET,SO_REUSEADDR,
                                       (char*)&k,sizeof(int)) < 0) 
  {
  fatal_error("Failed setsockopt reuse");
  return FALSE;
  }
memset(&netrec_addr,0,sizeof(netrec_addr));
netrec_addr.sin_family=AF_INET;
netrec_addr.sin_addr.s_addr=htonl(INADDR_ANY);
lir_text(0,1,"Listening to network for FFT2 input");
portno=ui.base_port+FFT2_PORT_OFFSET;
sprintf(s,"Group [%s]  port [%d] ", netrec_rx_multi_group, portno);
lir_text(0,2,s);
lir_sched_yield();
lir_refresh_screen();
netrec_addr.sin_port=htons(portno);
// bind to receive address
if (bind(netfd_fft2in,(struct sockaddr *) &netrec_addr,
                                                sizeof(netrec_addr)) < 0) 
  {
  fatal_error("Failed bind");
  return FALSE;
  }
// use setsockopt() to request that the kernel join a multicast group
mreq.imr_multiaddr.s_addr=inet_addr(netrec_rx_multi_group);
mreq.imr_interface.s_addr=htonl(INADDR_ANY);
i=setsockopt(netfd_fft2in,IPPROTO_IP,IP_ADD_MEMBERSHIP,
                                       (char*)&mreq,sizeof(mreq));
if(i != 0)
  {
#if(OSNUM == OS_FLAG_WINDOWS)
  i=WSAGetLastError();
  sprintf(errms,"socket(...) failed with code %d",i);
#endif
#if(OSNUM == OS_FLAG_LINUX)
  fatal_error("Failed setsockopt join");
#endif
  return FALSE;   
  }
// Go into an endless loop waiting for multicast to be received
// from another Linrad system.
nbytes=0;
i=0;
test_keyboard();
FD_ZERO(&fds);
FD_SET(netfd_fft2in, &fds);
while(nbytes <= 0 && lir_inkey == 0)
  {
  if(kill_all_flag)return FALSE;
  testfds=fds;
  timeout.tv_sec=0;
  timeout.tv_usec=30000;
  if( select(FD_SETSIZE, &testfds, (fd_set *)0,  (fd_set *)0, &timeout) > 0)
    {   
    if(FD_ISSET(netfd_fft2in,&testfds))
      {
      memset(&netmaster_addr,0,sizeof(netmaster_addr));
      addrlen=sizeof(netmaster_addr);
      nbytes=recvfrom(netfd_fft2in,(char*)msg,sizeof(NET_RX_STRUCT)+1,0, 
                        (struct sockaddr *) &netmaster_addr, &addrlen);
      if(nbytes != sizeof(NET_RX_STRUCT))
        {
        i++;
        sprintf(s,"NET RX ERROR %d    received %d expected %d",i,
                                                nbytes,sizeof(NET_RX_STRUCT));
        lir_text(1,4,s);
        nbytes=0;
        }
      lir_sched_yield();  
      lir_refresh_screen();
      }
    }  
  else
    {
    lir_sleep(100000);
    test_keyboard();
    }          
  }    
setupflag=3;
setupcount=0;
FD_ZERO(&rx_input_fds);
FD_SET(netfd_fft2in, &rx_input_fds);
while(setupflag != 0)
  {
  setupcount++;
  if( setupcount > 4)
    {
    fatal_error("Failed to get FFT2 parameters");
    return FALSE;
    }
  if( (msg->block_no&1) == 0)
    {
// block_no is even.
    setupflag&=-2;
// userx_no is the number of receiver RF channels.
// Reversed sign indicates float format.    
    if(msg->userx_no < 0)
      {
      no_of_rx_channels=-msg->userx_no;
      swfloat=TRUE;
      }
    else
      {
      no_of_rx_channels=msg->userx_no;
      swfloat=FALSE;
      }
    timf2_sampling_speed=msg->userx_freq;
    }  
  else
    {
// block_no is odd.
    setupflag&=-3;
    fft2_n=msg->userx_no;
    fft2_sinpow=msg->userx_freq;
    }
  lir_netread_rxin(msg);
  }
if(lir_inkey != 0)
  {
  lir_inkey=0;
  fatal_error("Wait interrupted from keyboard");
  return FALSE;
  } 
old_passband_center=msg->passband_center;
passband_center=msg->passband_center;
ip=(void*)&netmaster_addr.sin_addr.s_addr;
sprintf(s,"received FFT2 data from  %d.%d.%d.%d",ip[0], ip[1], ip[2], ip[3]);
clear_screen();
lir_text(0,0,s);
lir_refresh_screen();
return TRUE;
}

