// For help on X11 give command: "man X Interface"
// Event masks and event definitions are in /usr/X11R6/include/X11/X.h


#include <string.h>
#include <X11/Xlib.h>
#include <ctype.h>
#include <semaphore.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>
#include <X11/cursorfont.h>
#include <X11/Xutil.h>
#include "globdef.h"
#include "vernr.h"
#include "mousedef.h"
#include "keyboard_def.h"
#include "fontdef.h"
#include "uilim.h"

extern int netfd_fft2in;

GC xgc;
XImage *ximage;
Display *xdis;
Window xwin;
Colormap lir_colormap;
int expose_event_done;
unsigned char *mempix_char;
unsigned short int *mempix_shi;
int first_mempix;
int last_mempix;
int shift_key_status;
int color_depth;
int process_event_flag;
unsigned char *xpalette;

char mouse_thread_flag;

pthread_t thread_identifier_kill_all;
pthread_t thread_identifier_keyboard;
pthread_t thread_identifier_mouse;
pthread_t thread_identifier_process_event;
pthread_t thread_identifier_zoom;
pthread_t thread_identifier_fft2;

sem_t sem_kill_all;
sem_t sem_mouse;
sem_t sem_keyboard;

void ui_setup(void);
void thread_kill_all(void);
void thread_process_event(void);
void thread_mouse(void);
void lir_move_mouse_cursor(void);
void lir_remove_mouse_thread(void);
void store_in_kbdbuf(int c);
void thread_zoom(void);
void zoom(void);


typedef struct {
unsigned short int red;
unsigned short int green;
unsigned short int blue;
unsigned int pixel;
short int flag;
float total;
}PIXINFO;



int main(int argc, char **argv)
{
int id,il,kd,kl;
int bitmap_pad;
float t1,t2;
PIXINFO *defpix, *lirpix;
float *pixdiff;
PIXINFO tmppix;
Colormap default_colormap;
char *hostname;
Visual *visual;
int i, k, m, screen_num;
Cursor cross_cursor;
char *window_name=PROGRAM_NAME;
XTextProperty window_title_property;
unsigned short int *ipalette;
XColor xco;
XInitThreads();
if(DUMPFILE)
  {
  dmp = fopen("dmp", "w");
  DEB"\n******************************\n");
  }
else
  {  
  dmp=NULL;
  }
os_flag=OS_FLAG_X;
expose_event_done=FALSE;
first_mempix=0x7fffffff;
last_mempix=0;
shift_key_status=0;
i=argc;
keyboard_buffer_ptr=0;
keyboard_buffer_used=0;
keyboard_buffer=malloc(KEYBOARD_BUFFER_SIZE*sizeof(int));
if (argv[1] == NULL)
  {
  hostname = NULL;
  }
else
  {
  hostname = argv[1];
  }
xdis = XOpenDisplay(hostname);
if (xdis == NULL)
  {
  fprintf(stderr, "\nCan't open display: %s\n", hostname);
  return(10);
  }
visual=DefaultVisual(xdis, 0);
ui_setup();
screen_num = DefaultScreen(xdis);
screen_width = ui.screen_width_factor*DisplayWidth(xdis, screen_num)/100;
screen_width &= -4;
screen_height = ui.screen_height_factor*DisplayHeight(xdis, screen_num)/100;
screen_totpix=screen_width*(screen_height+1);
// *****************************************************************
// Set the variables Linrad uses to access the screen.
error_exit_msg=init_font(ui.font_scale);
if(error_exit_msg != NULL)goto exitmain;
xwin=XCreateSimpleWindow(xdis, RootWindow(xdis, 0), 
                                    0, 0, screen_width, screen_height, 1, 0, 0);
cross_cursor = XCreateFontCursor(xdis, XC_diamond_cross);
// attach the icon cursor to our window
XDefineCursor(xdis, xwin, cross_cursor);
// Set the name of our window
// This is the string to be translated into a property. 
// translate the given string into an X property. 
XStringListToTextProperty(&window_name, 1, &window_title_property);
XSetWMName(xdis, xwin, &window_title_property);
xgc=DefaultGC(xdis, 0);
color_depth = DefaultDepth(xdis, screen_num );
if(visual->class!=TrueColor && color_depth != 8)
  {
  printf("Unknown color type\n");
  exit(1);
  }
bitmap_pad=color_depth;  
switch (color_depth)
  {
  case 24:
  bitmap_pad=32;
  mempix_char=(unsigned char *)malloc((screen_totpix+1)*4);
  for(i=0; i<(screen_totpix+1)*4;i++)mempix_char[i]=0;
// ******************************************************************
// Rearrange the palette. It was designed for svgalib under Linux
  for(i=0; i<3*256; i++)
    {
    svga_palette[i]<<=2;
    if(svga_palette[i] != 0) svga_palette[i]|=3;
    }
  break;
  
  case 16:
  mempix_shi=(unsigned short int*)malloc((screen_totpix+1)*sizeof(unsigned short int));
  mempix_char=(void*)mempix_shi;
  for(i=0; i<screen_totpix+1; i++)mempix_shi[i]=0;
// ******************************************************************
  ipalette=(void*)(&svga_palette[0]);
  for(i=0; i<256; i++)
    {
    k=svga_palette[3*i+2];
    k&=0xfffe;
    k<<=5;
    k|=svga_palette[3*i+1];
    k&=0xfffe;
    k<<=6;
    k|=svga_palette[3*i  ];
    k>>=1;
    ipalette[i]=k;
    }
  break;
  
  case 8:
  mempix_char=(unsigned char*)malloc((screen_totpix+1)+256);
  defpix=(PIXINFO*)malloc(sizeof(PIXINFO)*256);
  lirpix=(PIXINFO*)malloc(sizeof(PIXINFO)*256);
  pixdiff=(float*)malloc(256*256*sizeof(float));
  xpalette=&mempix_char[screen_totpix+1];
  for(i=0; i<(screen_totpix+1)+256; i++)mempix_char[i]=0;
  lir_colormap=XCreateColormap(xdis, xwin, visual, AllocAll);
  default_colormap = DefaultColormap(xdis, screen_num);
// Store the default colormap in defpix
  for(id=0; id<256; id++)
    {
    xco.pixel=id;
    k=XQueryColor (xdis,default_colormap,&xco);  
    defpix[id].red=xco.red;
    defpix[id].green=xco.green;
    defpix[id].blue=xco.blue;
    defpix[id].flag=0;
    defpix[id].pixel=id;
    defpix[id].total=    ( (float)((unsigned int)defpix[id].red)*
                                  (unsigned int)defpix[id].red+
                          (float)((unsigned int)defpix[id].green)*
                                  (unsigned int)defpix[id].green+
                          (float)((unsigned int)defpix[id].blue)*
                                  (unsigned int)defpix[id].blue);
    }
// svga_palette uses the six lowest bits for the colour intensities.
// shift left by 10 to move our data to occupy the six highest bits.
// Store the svgalib palette.
  for(il=0; il<MAX_SVGA_PALETTE; il++)
    {
    lirpix[il].red=svga_palette[3*il+2]<<2;
    lirpix[il].green=svga_palette[3*il+1]<<2;
    lirpix[il].blue=svga_palette[3*il  ]<<2;
    if(lirpix[il].red != 0)lirpix[il].red|=3;
    if(lirpix[il].green != 0)lirpix[il].green|=3;
    if(lirpix[il].blue != 0)lirpix[il].blue|=3;
    lirpix[il].red<<=8;
    lirpix[il].green<<=8;
    lirpix[il].blue<<=8;
    lirpix[il].pixel=il;
    lirpix[il].flag=1;
    lirpix[il].total=   ( (float)((unsigned int)lirpix[il].red)*
                                  (unsigned int)lirpix[il].red+
                          (float)((unsigned int)lirpix[il].green)*
                                  (unsigned int)lirpix[il].green+
                          (float)((unsigned int)lirpix[il].blue)*
                                  (unsigned int)lirpix[il].blue);
    }
  for(il=MAX_SVGA_PALETTE; il<256; il++)
    {
    lirpix[il].red=0;
    lirpix[il].green=0;
    lirpix[il].blue=0;
    lirpix[il].pixel=il;
    lirpix[il].flag=0;
    }
#define M 0.00000001
#define N 0x100
// Sort lirpix in order of ascending total intensity.
  for(il=0; il<MAX_SVGA_PALETTE-1; il++)
    {
    t1=0;
    m=il;
    for(kl=il; kl<MAX_SVGA_PALETTE; kl++)
      {
      if(lirpix[kl].total > t1)
        {
        t1=lirpix[kl].total;
        m=kl;
        }
      }  
    tmppix=lirpix[il];
    lirpix[il]=lirpix[m];
    lirpix[m]=tmppix;
    }
// Compute the similarity between lirpix and defpix and store in a matrix.
  for(il=0; il<MAX_SVGA_PALETTE; il++)
    {
    for(id=0; id<256; id++)
      {
      t2=pow((float)(int)(((unsigned int)lirpix[il].red-
                      (unsigned int)defpix[id].red)),2.0)+
         pow((float)(int)(((unsigned int)lirpix[il].green-
                      (unsigned int)defpix[id].green)),2.0)+
         pow((float)(int)(((unsigned int)lirpix[il].blue-
                      (unsigned int)defpix[id].blue)),2.0);
      pixdiff[id+il*256]=t2;
      }
    }  
// Reorder the default colormap for the diagonal elements
// of pixdiff (up to MAX_SVGA_PALETTE-1) to become as small
// as possible when stepping in the ascending order that
// lirpix currently is sorted in.
  for(il=0; il<MAX_SVGA_PALETTE; il++)
    {
    t1=BIG;
    kd=0;
    for(id=0; id<256; id++)
      {
      if(pixdiff[id+il*256] < t1)
        {
        t1=pixdiff[id+il*256];
        kd=id;
        }
      }
    tmppix=defpix[il];
    defpix[il]=defpix[kd];
    defpix[kd]=tmppix;
    for(kl=0; kl<MAX_SVGA_PALETTE; kl++)
      {
      t1=pixdiff[kd+kl*256];
      pixdiff[kd+kl*256]=pixdiff[il+kl*256];
      pixdiff[il+kl*256]=t1;
      }
    }
  for(i=0; i<MAX_SVGA_PALETTE; i++)
    {
    xco.pixel=defpix[i].pixel;
    xco.red=lirpix[i].red;
    xco.green=lirpix[i].green;
    xco.blue=lirpix[i].blue;
    xco.flags=DoRed|DoGreen|DoBlue;
    xco.pad=0;
    k=XStoreColor(xdis, lir_colormap, &xco);  
    if(k==0)
      {
      printf("\nPalette failed\n");
      goto exitmain;
      }
    xpalette[lirpix[i].pixel]=xco.pixel;
    }
  for(i=MAX_SVGA_PALETTE; i<256; i++)
    {
    xco.pixel=defpix[i].pixel;
    xco.red=defpix[i].red;
    xco.green=defpix[i].green;
    xco.blue=defpix[i].blue;
    xco.flags=DoRed|DoGreen|DoBlue;
    xco.pad=0;
    k=XStoreColor(xdis, lir_colormap, &xco);  
    if(k==0)
      {
      printf("\nPalette failed\n");
      goto exitmain;
      }
    xpalette[i]=lirpix[i].pixel;
    }
  XSetWindowColormap(xdis, xwin, lir_colormap);
  free(defpix);
  free(lirpix);
  free(pixdiff);
  break;
  
  default:
  printf("\nUnknown color depth: %d\n",color_depth);
  goto exitmain;
  } 
ximage=XCreateImage(xdis, visual, color_depth, ZPixmap, 0, 
          (char*)(mempix_char), screen_width, screen_height, bitmap_pad, 0);
XInitImage(ximage);
XSelectInput(xdis, xwin, 
             ButtonPressMask|
             ExposureMask|
             KeyPressMask|
             KeyReleaseMask| 
             ButtonReleaseMask|
             PointerMotionMask);
XMapWindow(xdis, xwin);

// Create a thread that will close the entire process in a controlled way
// in case lirerr() is called or the ESC key is pressed.
sem_init(&sem_kill_all,0,0);
kill_all_flag=FALSE;
pthread_create(&thread_identifier_kill_all,NULL,(void*)thread_kill_all, NULL);
sem_init(&sem_keyboard,0,0);
process_event_flag=TRUE;
pthread_create(&thread_identifier_process_event,NULL,
                                        (void*)thread_process_event, NULL);
while(expose_event_done == FALSE)
  {
  usleep(1000);
  }
sem_init(&sem_mouse,0,0);
pthread_create(&thread_identifier_mouse,NULL, (void*)thread_mouse, NULL);
// Create a thread for the zoom program.
if(kill_all_flag) goto skipzoom;
pthread_create(&thread_identifier_zoom,NULL, (void*)thread_zoom, NULL);
pthread_join(thread_identifier_zoom,0);
skipzoom:;
sem_post(&sem_kill_all);
pthread_join(thread_identifier_kill_all,0);
lir_remove_mouse_thread();
pthread_join(thread_identifier_mouse,0);
lir_refresh_screen();
XSync(xdis,True);
process_event_flag=FALSE;
pthread_join(thread_identifier_process_event,0);
XFreeCursor(xdis, cross_cursor);
free(mempix_char);
exitmain:;
free(vga_font);
XCloseDisplay(xdis);
free(keyboard_buffer);
if(dmp!=NULL)fclose(dmp);
if(error_exit_msg != NULL)printf("\n%s",error_exit_msg);
return 0;
}

void thread_process_event(void)
{
int chr;
int cc;
XEvent ev;
char key_buff[16];
int count;
KeySym ks;
while(process_event_flag)
  {
  XNextEvent(xdis, &ev);
  switch(ev.type)
    {
    case Expose:
    XPutImage(xdis, xwin, xgc, ximage,0,0,0,0, screen_width, screen_height);
    if(color_depth==8)XInstallColormap(xdis, lir_colormap);
    expose_event_done=TRUE;
    break;

    case ButtonPress:
    if ( (ev.xbutton.button == Button1) != 0)
      {
      new_lbutton_state=1;
      goto mousepost;
      }
    if ( (ev.xbutton.button == Button3) != 0) 
      {
      new_rbutton_state=1;
      goto mousepost;
      }
    break;

    case ButtonRelease:
    if ( (ev.xbutton.button == Button1) != 0)
      {
      new_lbutton_state=0;
      goto mousepost;
      }
    if ( (ev.xbutton.button == Button3) != 0) 
      {
      new_rbutton_state=0;
      goto mousepost;
      }
    break;

    case MotionNotify:
    new_mouse_x= ev.xbutton.x;
    if(new_mouse_x < 0)new_mouse_x=0;
    if(new_mouse_x >= screen_width)new_mouse_x=screen_width-1;
    new_mouse_y= ev.xbutton.y;
    if(new_mouse_y < 0)new_mouse_y=0;
    if(new_mouse_y >= screen_height)new_mouse_y=screen_height-1;
mousepost:;
    sem_post(&sem_mouse);
    break;
  
    case KeyPress:
    chr = XKeycodeToKeysym(xdis, ev.xkey.keycode, 0);
    if(chr == X_ESC_SYM)
      {
// The ESC key was pressed. Exit from Linrad NOW!    
      sem_post(&sem_kill_all);
      store_in_kbdbuf(0);
      return;
      }    
    if(chr == X_SHIFT_SYM_L || chr == X_SHIFT_SYM_R)
      {
      shift_key_status=1;
      break;
      }
// Get ASCII codes from Dec 32 to Dec 127
    count = XLookupString((XKeyEvent *)&ev, key_buff, 2, &ks,NULL);
    if ((count == 1) && ((int)key_buff[0]>31) )
      {
      store_in_kbdbuf((int)key_buff[0]);
      break;
      }
    cc=0;
    if(chr >= X_F1_SYM && chr <= X_F12_SYM)
      {
      if(chr <= X_F10_SYM)
        {
        if(shift_key_status == 0)
          {
          cc=F1_KEY+chr-X_F1_SYM;
          }
        else  
          {
          cc=SHIFT_F1_KEY+chr-X_F1_SYM;
          if(chr > X_F2_SYM)cc++;
          if(chr >= X_F9_SYM)cc=0;
          }
        }
      else
        {
        if(shift_key_status == 0)
          {
          cc=F11_KEY+chr-X_F11_SYM;
          }
        }  
      if(cc != 0)store_in_kbdbuf(cc);        
      break;
      }
    switch (chr)
      {
      case X_UP_SYM:
      cc=ARROW_UP_KEY;
      break;

      case X_DWN_SYM:
      cc=ARROW_DOWN_KEY;
      break;

      case X_RIGHT_SYM:
      cc=ARROW_RIGHT_KEY;
      break;

      case  X_BACKDEL_SYM:
      case X_LEFT_SYM:
      cc=ARROW_LEFT_KEY;
      break;
      
      case X_HOME_SYM:
      cc=HOME_KEY;
      break;
      
      case X_INSERT_SYM:
      cc=INSERT_KEY;
      break;
      
      case X_DELETE_SYM:
      cc=DELETE_KEY;
      break;
      
      case X_END_SYM:
      cc=END_KEY;
      break;
      
      case X_PGUP_SYM:
      cc=PAGE_UP_KEY;
      break;

      case X_PGDN_SYM:
      cc=PAGE_DOWN_KEY;
      break;
      
      case X_PAUSE_SYM:
      cc=PAUSE_KEY;
      break;
      
      case X_ENTER_SYM:
      cc=10;
      break;

      }
    if(cc != 0)store_in_kbdbuf(cc);        
    break;

    case KeyRelease:
    chr=XKeycodeToKeysym(xdis, ev.xkey.keycode, 0);
    if(chr == X_SHIFT_SYM_L || chr == X_SHIFT_SYM_R)
      {
      shift_key_status=0;
      }
    break;        
    }
  }
}

void ui_setup(void)
{
FILE *file;
int i, j, k, nread;
char s[10];
char chr;
int *uiparm;
char *parinfo;
char *errms;
uiparm=(int*)(&ui);
parinfo=NULL;
file = fopen(userint_filename, "rb");
if (file == NULL)
  {
  printf("\nSetup file %s missing.",userint_filename);
full_setup:;
  for(i=0; i<MAX_UIPARM; i++) uiparm[i]=0;
  printf("\nForcing setup to create a new %s file.",userint_filename);
  printf("\n\nPress S to continue");
  printf("\nAny other key to exit."); 
  printf("\nThen press enter\n\n=>"); 
  fgets(s,8,stdin);
  chr=toupper(s[0]);
  if(chr != 'S') exit(0);
  if(x_global_uiparms(0) == FALSE)exit(0);
  }
else
  {
  if(parinfo==NULL)
    {
    parinfo=malloc(4096);
    if(parinfo==NULL)exit(0);
    }
  for(i=0; i<4096; i++) parinfo[i]=0;
  nread=fread(parinfo,1,4095,file);
  fclose(file);
  file=NULL;
  if(nread >= 4095)
    {
    goto full_setup;
    }
  k=0;
  for(i=0; i<MAX_UIPARM; i++)
    {
    while(parinfo[k]==' ' ||
          parinfo[k]== '\n' )
      {
      k++;
      if(k>=nread-1)goto full_setup;
      }
    j=0;
    while(parinfo[k]== uiparm_text[i][j] && k<nread)
      {
      k++;
      j++;
      } 
    if(uiparm_text[i][j] != 0)goto full_setup;
    while(parinfo[k]!='[' && k<4096)k++;
    sscanf(&parinfo[k],"[%d]",&uiparm[i]);
    while(parinfo[k]!='\n')k++;
    }
  free(parinfo);
  if(ui.font_scale < 1 || ui.font_scale >5)
    {
    printf("\nIllegal paramerer value in setup file %s",userint_filename);
    goto full_setup;
    }
  if(ui.screen_width_factor < MIN_SCREEN_W || 
     ui.screen_width_factor > MAX_SCREEN_W ||
     ui.screen_height_factor < MIN_SCREEN_H || 
     ui.screen_height_factor > MAX_SCREEN_H)
    {
    printf("\nIllegal screen scale factors.");
    goto full_setup;
    }
  if(ui.max_filesiz < MIN_FILESIZ || ui.max_filesiz > MAX_FILESIZ)
    {
    printf("\nIllegal value for max scratch file size");
    goto full_setup;
    }
  if(ui.base_port < NET_MULTI_MINPORT || ui.base_port > NET_MULTI_MAXPORT)
    {
    printf("\nBase port out of range");    
    goto full_setup;
    }
  if(ui.base_port!=(ui.base_port/10)*10)
    {
    printf("\nIllegal base port. Must be a multiple of 10.");
    goto full_setup;
    }
  if(ui.netaddr_offset <0 || ui.netaddr_offset > 15)
    {
    printf("\nNet address offset out of range");
    goto full_setup;
    }  
  if(ui.max_lines < 50)
    {
    printf("\nMax lines too low");
    goto full_setup;
    }  
  }
errms=save_uiparm();
if(errms != NULL)
  {
  printf("\n%s",errms);
  exit(0);
  }
}

void thread_kill_all(void)
{
// Wait until the semaphore is released.
// Then stop all processing threads so main can write any
// error message and exit.
sem_wait(&sem_kill_all);
kill_all_flag=TRUE;
}

void lir_remove_mouse_thread(void)
{
if(mouse_thread_flag == THRFLAG_NOT_ACTIVE)return;
mouse_thread_flag=THRFLAG_KILL;
while(mouse_thread_flag != THRFLAG_RETURNED)
  {
  sem_post(&sem_mouse);
  usleep(1000);
  }
}

void store_in_kbdbuf(int c)
{
if(c==13)c=10;
keyboard_buffer[keyboard_buffer_ptr]=c;
keyboard_buffer_ptr=(keyboard_buffer_ptr+1)&(KEYBOARD_BUFFER_SIZE-1);
sem_post(&sem_keyboard);
}

// *********************************************************
// Mouse routines for Windows and X11. Draw nothing on screen.
// *********************************************************

void unconditional_hide_mouse(void)
{
// Do nothing, Windows or X11 take care of the mouse cursor
}

void show_mouse(void)
{
// Do nothing, Windows or X11 take care of the mouse cursor
}


void hide_mouse(int x1, int x2, int iy1, int y2)
{
int i;
i=x1|x2|iy1|y2;
// Do nothing, Windows or X11 take care of the mouse cursor
}

void lir_move_mouse_cursor(void)
{
set_button_coordinates();
// We do not have a cursor of our own. Do nothing.
}

void thread_mouse(void)
{
mouse_thread_flag=THRFLAG_ACTIVE;
while(mouse_thread_flag==THRFLAG_ACTIVE)
  {
// **********************************************
// Wait until a mouse event occurs. 
  sem_wait(&sem_mouse);
  lir_move_mouse_cursor();
  check_mouse_actions();
  }
mouse_thread_flag=THRFLAG_RETURNED;  
}

int x_global_uiparms(int wn)
{
char *errms;
char s[80],ss[10];
int line;
line=2;
if(ui.vga_mode==0)
  {
  ui.vga_mode=10;
  if(screen_width > 640)ui.vga_mode=11;
  if(screen_width > 800)ui.vga_mode=12;
  }
//  ********************  FONT SIZE  ************************
sprintf(s,"Font scale (1 to 5):"); 
fntc:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  lir_inkey=toupper(ss[0]);
  ui.font_scale=lir_inkey-'0';
  }
else
  {
  clear_screen();
  lir_text(0,line,s);
  ui.font_scale=lir_get_integer(21,line,2,1,5);
  if(kill_all_flag)return FALSE;
  line++;
  }
if(lir_inkey < '1' || lir_inkey > '5')goto fntc;
// ******************** SCREEN WIDTH ************************  
sprintf(s,"Percentage of screen width to use(%d to %d):",
                                       MIN_SCREEN_W, MAX_SCREEN_W); 
wfac:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.screen_width_factor);
  if(ui.screen_width_factor < MIN_SCREEN_W ||
     ui.screen_width_factor > MAX_SCREEN_W)goto wfac;

  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.screen_width_factor=lir_get_integer(47,line,3,
                                            MIN_SCREEN_W,MAX_SCREEN_W);
  if(kill_all_flag) return FALSE;
  line++;
  }
// ******************** BASE PORT ************************  
sprintf(s,"Network base port (%d to %d) =>",
                                  NET_MULTI_MINPORT, NET_MULTI_MAXPORT);
basepo:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.base_port);
  if(ui.base_port < NET_MULTI_MINPORT  ||
     ui.base_port > NET_MULTI_MAXPORT)goto basepo;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.base_port=lir_get_integer(47,line,5,
                                 NET_MULTI_MINPORT, NET_MULTI_MAXPORT);
  if(kill_all_flag) return FALSE;
  line++;
  }
// ******************** GROUP LAST BYTE ************************  
sprintf(s,"[%s] New last byte (0 to 15)", netrec_rx_multi_group);
addr:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.netaddr_offset);
  if(ui.netaddr_offset < 0 || ui.netaddr_offset > 15)goto addr;  
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.netaddr_offset=lir_get_integer(60,line,2,0,15);
  if(kill_all_flag) return FALSE;
  line++;
  }
// ********************** SCREEN HEIGHT *********************** 
sprintf(s,"Percentage of screen height to use (%d to %d):",
                                            MIN_SCREEN_H, MAX_SCREEN_H); 
hfac:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.screen_height_factor);
  if(ui.screen_height_factor < MIN_SCREEN_H ||
     ui.screen_height_factor > MAX_SCREEN_H)goto hfac;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.screen_height_factor=lir_get_integer(49,line,3,MIN_SCREEN_H,MAX_SCREEN_H);
  if(kill_all_flag) return FALSE;
  line++;
  }
// ********************** SCRATCH FILE SIZE ********************  
sprintf(s,"Max scratch file size in powers of two(%d to %d):",
                                                  MIN_FILESIZ, MAX_FILESIZ);
filemax:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.max_filesiz);
  if(ui.max_filesiz < MIN_FILESIZ || ui.max_filesiz > MAX_FILESIZ)goto filemax;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.max_filesiz=lir_get_integer(49,line,3,MIN_FILESIZ,MAX_FILESIZ);
  if(kill_all_flag) return FALSE;
  line++;
  }
// ********************** WATERFALL LINES LIMITATION *************  
sprintf(s,"Waterfall lines limitation (50 and above)):");
linemax:;
if(wn==0)
  {
  printf("\n%s\n=>",s); 
  fgets(ss,8,stdin);
  sscanf(ss,"%d", &ui.max_lines);
  if(ui.max_lines < 50)goto linemax;
  }
else
  {
  clear_lines(line,line);
  lir_text(0,line,s);
  ui.max_lines=lir_get_integer(49,line,4,50,10000);
  if(kill_all_flag) return FALSE;
  }
// ******************************************************
errms=save_uiparm();
if(errms != NULL)
  {
  if(wn==0)
    {
    printf("\%s",errms);
    }
  else
    {
    lir_text(0,8,errms);
    lir_text(15,9,press_any_key);
    await_keyboard();
    }
  return FALSE;
  }    
return TRUE;
}

void clear_screen(void)
{
int i, k;
switch (color_depth)
  {
  case 24:
  k=4*screen_width*screen_height;
  for(i=0; i<k; i+=4)
    {
    mempix_char[i]=0;
    mempix_char[i+1]=0;
    mempix_char[i+2]=0;
    mempix_char[i+3]=255;
    }
  break;

  case 16:
  k=screen_width*screen_height;
  for(i=0; i<k; i++)
    {
    mempix_shi[i]=0;
    }
  break;
  
  case 8:
  k=screen_width*screen_height;
  for(i=0; i<k; i++)
    {
    mempix_char[i]=xpalette[0];
    }
  }
first_mempix=0;
last_mempix=screen_totpix-1;
lir_refresh_screen();
}

void lir_refresh_screen(void)
{
int l1, h, h1, n;
if(last_mempix >= first_mempix )
  {
  l1=first_mempix/screen_width;
  h=last_mempix/screen_width-l1+1;
  if(h>screen_height)h=screen_height;
  first_mempix=0x7fffffff;
  last_mempix=0;
  while(h>0)
    {
    n=(5*h)/screen_height;
    if(n > 1)
      {
      h1=h/n;
      }
    else
      {
      h1=h;
      }
    XPutImage(xdis, xwin, xgc, ximage, 0, l1, 0, l1, screen_width, h1);
    XFlush(xdis);
    l1+=h1;
    h-=h1;
    if(h > 0)sched_yield();
    }
  XFlush(xdis);
  }
}

void aw_keyb(void)
{
while(keyboard_buffer_ptr == keyboard_buffer_used)
  {
  if(keyboard_buffer_ptr == keyboard_buffer_used)
    {
    sem_wait(&sem_keyboard);
    if(kill_all_flag) return;
    }
  }
lir_inkey = keyboard_buffer[keyboard_buffer_used];
keyboard_buffer_used=(keyboard_buffer_used+1)&(KEYBOARD_BUFFER_SIZE-1);
}


void await_keyboard(void)
{
lir_refresh_screen();
aw_keyb();
}

void lir_putbox(int x, int y, int w, int h, void* dp)
{
unsigned char *savmem;
int i, j, k, m;
unsigned short int *ipalette;
k=(x+y*screen_width);
if(k<0 || k+(h-1)*screen_width+w-1 >= screen_totpix)
  {
  DEB"putbox ERROR");
  return;
  }
if(first_mempix > k)first_mempix=k;
m=0;
savmem=(unsigned char*)dp;
switch (color_depth)
  {
  case 24:
  for(j=0; j<h; j++)
    {  
    for(i=0; i<w; i++)
      {
      mempix_char[4*(k+i)  ]=svga_palette[3*savmem[m]  ];
      mempix_char[4*(k+i)+1]=svga_palette[3*savmem[m]+1];
      mempix_char[4*(k+i)+2]=svga_palette[3*savmem[m]+2];
      m++;
      }
    k+=screen_width;
    }
  break;
  
  case 16:
  ipalette=(void*)(&svga_palette[0]);
  for(j=0; j<h; j++)
    {  
    for(i=0; i<w; i++)
      {
      mempix_shi[k+i]=ipalette[savmem[m]];
      m++;
      }
    k+=screen_width;
    }
  break;

  case 8:
  for(j=0; j<h; j++)
    {  
    for(i=0; i<w; i++)
      {
      mempix_char[k+i]=xpalette[savmem[m]];
      m++;
      }
    k+=screen_width;
    }
  }  
if(last_mempix < k)last_mempix=k;
}

void lir_fillbox(int x, int y,int w, int h, unsigned char c)
{
int i, j, k;
unsigned short int *ipalette;
k=(x+y*screen_width);
if(k<0 || k+(h-1)*screen_height+w > screen_totpix)
  {
  DEB"fillbox ERROR");
  return;
  }
if(first_mempix > k)first_mempix=k;
switch (color_depth)
  {
  case 24:
  for(j=0; j<h; j++)
    {
    for(i=0; i<=w; i++)
      {
      mempix_char[4*(k+i)  ]=svga_palette[3*c  ];
      mempix_char[4*(k+i)+1]=svga_palette[3*c+1];
      mempix_char[4*(k+i)+2]=svga_palette[3*c+2];
      }
    k+=screen_width;
    }
  break;
      
  case 16:
  ipalette=(void*)(&svga_palette[0]);
  for(j=0; j<h; j++)
    {
    for(i=0; i<=w; i++)
      {
      mempix_shi[k+i]=ipalette[c];
      }
    k+=screen_width;
    }
  break;

  case 8:
  for(j=0; j<h; j++)
    {
    for(i=0; i<=w; i++)
      {
      mempix_char[k+i]=xpalette[c];
      }
    k+=screen_width;
    }
  break;
  }    
k-=screen_width-w;
if(last_mempix < k)last_mempix=k;
}

void lir_hline(int x1, int y, int x2, unsigned char c)
{
int i, ia, ib;
unsigned short int *ipalette;
ia=y*screen_width;
if(x1 <= x2)
  {
  ib=ia+x2;
  ia+=x1;
  }
else
  {
  ib=ia+x1;
  ia+=x2;
  }
if(ia < 0 || ib >= screen_totpix)return;
switch (color_depth)
  {
  case 24:
  for(i=ia; i<=ib; i++)
    {
    mempix_char[4*i  ]=svga_palette[3*c  ];
    mempix_char[4*i+1]=svga_palette[3*c+1];
    mempix_char[4*i+2]=svga_palette[3*c+2];
    }
  break;
  
  case 16:
  ipalette=(void*)(&svga_palette[0]);
  for(i=ia; i<=ib; i++)
    {
    mempix_shi[i]=ipalette[c];
    }
  break;

  case 8:
  for(i=ia; i<=ib; i++)
    {
    mempix_char[i]=xpalette[c];
    }
  break;
  }
if(first_mempix > ia)first_mempix=ia;
if(last_mempix < ib)last_mempix=ib;
}

void lir_line(int x1, int yy1, int x2,int y2, unsigned char c)
{
int ia;
int i,j,k;
int xd, yd;
float t1,t2,delt;
unsigned short int *ipalette;
if(x1 < 0)x1=0;
if(x1 >= screen_width)x1=screen_width-1;
if(x2 < 0)x2=0;
if(x2 >= screen_width)x2=screen_width-1;
if(yy1 < 0)yy1=0;
if(yy1 >= screen_height)yy1=screen_height-1;
if(y2 < 0)y2=0;
if(y2 >= screen_height)y2=screen_height-1;
xd=x2-x1;
yd=y2-yy1;
if(yd==0)
  {
  if(xd==0)
    {
    lir_setpixel(x1,yy1,c);
    }
  else
    {
    lir_hline(x1,yy1,x2,c);
    }
  return;
  }  
if(abs(xd)>=abs(yd))
  {
  if(xd>=0)
    {
    k=1;
    }
  else
    {
    k=-1;
    }  
  if(yd >= 0)
    {
    delt=0.5;
    }
  else
    {  
    delt=-0.5;
    }
  t1=yy1;
  t2=(float)(yd)/abs(xd);
  i=x1-k;
  switch (color_depth)
    {
    case 24:
    while(i!=x2)
      {
      i+=k;
      j=t1+delt;
      ia=i+j*(screen_width);
      mempix_char[4*ia  ]=svga_palette[3*c  ];
      mempix_char[4*ia+1]=svga_palette[3*c+1];
      mempix_char[4*ia+2]=svga_palette[3*c+2];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;
    
    case 16:
    ipalette=(void*)(&svga_palette[0]);
    while(i!=x2)
      {
      i+=k;
      j=t1+delt;
      ia=i+j*(screen_width);
      mempix_shi[ia  ]=ipalette[c  ];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;
    
    case 8:
    while(i!=x2)
      {
      i+=k;
      j=t1+delt;
      ia=i+j*(screen_width);
      mempix_char[ia  ]=xpalette[c  ];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;
        
    }  
  }  
else
  {
  if(yd>=0)
    {
    k=1;
    }
  else
    {
    k=-1;
    } 
  if(xd >= 0)
    {
    delt=0.5;
    }
  else
    {  
    delt=-0.5;
    }
  t1=x1;
  t2=(float)(xd)/abs(yd);
  i=yy1-k;
  switch (color_depth)
    {
    case 24:
    while(i!=y2)
      {
      i+=k;
      j=t1+delt;
      ia=j+i*(screen_width);
      mempix_char[4*ia  ]=svga_palette[3*c  ];
      mempix_char[4*ia+1]=svga_palette[3*c+1];
      mempix_char[4*ia+2]=svga_palette[3*c+2];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;
    
    case 16:
    ipalette=(void*)(&svga_palette[0]);
    while(i!=y2)
      {
      i+=k;
      j=t1+delt;
      ia=j+i*(screen_width);
      mempix_shi[ia  ]=ipalette[c  ];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;

    case 8:
    while(i!=y2)
      {
      i+=k;
      j=t1+delt;
      ia=j+i*(screen_width);
      mempix_char[ia  ]=xpalette[c  ];
      if(first_mempix > ia)first_mempix=ia;
      if(last_mempix < ia)last_mempix=ia;
      t1+=t2;
      }
    break;
    }    
  }  
}

void lir_setpixel(int x, int y, unsigned char c)
{
int ia;
unsigned short int *ipalette;
ia=x+y*screen_width;
if(ia < 0 || ia >= screen_totpix)
  {
  DEB"\nwrite setpixel error: x=%d   y=%d",x,y);
  return;
  }
switch (color_depth)
  {
  case 24:
  mempix_char[4*ia  ]=svga_palette[3*c  ];
  mempix_char[4*ia+1]=svga_palette[3*c+1];
  mempix_char[4*ia+2]=svga_palette[3*c+2];
  break;
  
  case 16:
  ipalette=(void*)(&svga_palette[0]);
  mempix_shi[ia]=ipalette[c];
  break;

  case 8:
  mempix_char[ia]=xpalette[c];
  break;
  }
if(first_mempix > ia)first_mempix=ia;
if(last_mempix < ia)last_mempix=ia;
}


void test_keyboard(void)
{
if(kill_all_flag!=0)return;
if(keyboard_buffer_ptr != keyboard_buffer_used)
  {
  sem_wait(&sem_keyboard);
  lir_inkey = keyboard_buffer[keyboard_buffer_used];
  keyboard_buffer_used=(keyboard_buffer_used+1)&(KEYBOARD_BUFFER_SIZE-1);
  }
else
  {
  lir_inkey=0;  
  }
}

void thread_zoom(void)
{
netfd_fft2in=-1;
zoom();
if(netfd_fft2in != -1)close(netfd_fft2in);
}

void fatal_error(char *errmsg)
{
char *errms;
if(kill_all_flag) return;
errms=(void*)font8x12;
sprintf(errms,"FATAL ERROR:[%s]",errmsg);
DEB"\n%s",errms);
if(dmp != 0)fflush(dmp);
error_exit_msg=errms;
sem_post(&sem_kill_all);
usleep(100000);
}

void lir_sched_yield(void)
{
sched_yield();
}

void lir_sleep(int us)
{
usleep(us);
}

void create_netrec_fft2_thread(void)
{
pthread_create(&thread_identifier_fft2,NULL,
                              (void*)thread_rx_fft2_netinput, NULL);
}

void wait_for_netrec_exit(void)
{
pthread_join(thread_identifier_fft2,0);
}

void xz(char *s)
{
if(dmp == NULL) return;
PERMDEB" %s\n",s);
fflush(dmp);
sync();
}
