
#include <unistd.h>
#include "ctype.h"
#include "globdef.h"
#include "mousedef.h"
#include "fft2def.h"
#include "fontdef.h"
#include "keyboard_def.h"
#include "vernr.h"

void update_disk(void);

double zoom_freq;
int zoom_width_points;
int zoom_width_pixels;

void create_netrec_fft2_thread(void);
void wait_for_netrec_exit(void);
int open_network(void);


// block_pa  The latest block on scratch files.
// block_pw  The starting point for the current waterfall if not real time.
// block_pb  The latest block up to which average powers are computed.
// block_px  The oldest block containing valid data.

int help_flag;

unsigned int file_size;
int blocks_per_file;
int block_pa;
int block_pb;
int block_px;
int block_pw;
int block_mask;
int file_mask;
int block_bytes;
int block_totsize;
int waterf_last_line;
int no_of_waterf_lines;
int waterf_avgnum;
int points_per_pixel;
int pixels_per_point;
int waterf_first_line;
int waterf_scroll;
double waterf_hz_per_point;
double waterf_hz_per_pixel;
int fft2_first_point;
int fft2_last_point;
int waterf_first_pixel;
int waterf_last_pixel;
int waterf_totpix;
int pwr_pa;
int pwr_pp;
int pwr_px;
int pwr_totsize;
int real_time_flag;


float waterf_cfac=20;
float waterf_czer;

int waterf_line;

#define NO_OF_TMPFILES 16
char *tmpfile_name[NO_OF_TMPFILES]={"$01","$02","$03","$04","$05","$06",
                "$07","$08","$09","$10","$11","$12","$13","$14","$15","$16"};
FILE *tmpfile_fd[NO_OF_TMPFILES];

#define ZOOM_FREQ 0
#define ZOOM_HZ_PER_PIXEL 1
#define ZOOM_AVGNUM 2
#define ZOOM_COLOUR_ZERO 3
#define ZOOM_COLOUR_FAC 4
#define ZOOM_DELAY_TIME 5
#define ZOOM_POL_MODE 6
#define ZOOM_MAX_SCRO 7


#define SIN30  0.5
#define SIN60  0.866025403784439
#define COS60  0.5
#define COS30  0.866025403784439

#define POL_LIN_0 0
#define POL_LIN_30 1
#define POL_LIN_60 2
#define POL_LIN_90 3
#define POL_LIN_120 4
#define POL_LIN_150 5
#define POL_LEFT 6
#define POL_RIGHT 7
#define POL_TOTAL 8
#define POL_MAX_TYPES 9

int pol_mode=6;
char *pol_mode_text[POL_MAX_TYPES]={"LIN 0",      //0
                     "LIN 30",     //1
                     "LIN 60",     //2
                     "LIN 90",     //3
                     "LIN 120",    //4
                     "LIN 150",    //5
                     "C LEFT",     //6 
                     "C RIGHT",    //7 
                     "TOTAL"};     //8 

float get_power(int ptr)
{
if(no_of_rx_channels==1)return fft2_powersum[ptr];  
switch (pol_mode)
  {
  case POL_LIN_0:
  return fft2_xysum[ptr].lin0;
  break;
  
  case POL_LIN_30:
  return fft2_xysum[ptr].lin30;
  break;

  case POL_LIN_60:
  return fft2_xysum[ptr].lin60;
  break;

  case POL_LIN_90:
  return fft2_xysum[ptr].total-fft2_xysum[ptr].lin0;
  break;

  case POL_LIN_120:
  return fft2_xysum[ptr].total-fft2_xysum[ptr].lin30;
  break;

  case POL_LIN_150:
  return fft2_xysum[ptr].total-fft2_xysum[ptr].lin60;
  break;

  case POL_LEFT:
  return fft2_xysum[ptr].circ;
  break;

  case POL_RIGHT:
  return fft2_xysum[ptr].total-fft2_xysum[ptr].circ;
  break;

  case POL_TOTAL:
  return fft2_xysum[ptr].total;
  break;
  }
return 1;
}  

void draw_waterf_line(int line,int ptr)
{
int i, k, ia, ib, ix;
float t1, t2, yval,der;
int color;
if(help_flag == TRUE)return;
if(points_per_pixel == 1 || pixels_per_point == 1)
  {
  i=fft2_first_point;
  for(ix=waterf_first_pixel; ix < waterf_last_pixel; ix++)
    {
    t1=log10(get_power(ptr+i));
    t1=waterf_cfac*(t1-waterf_czer);
    if(t1<0)t1=0;
    if(t1 >= MAX_COLOR_SCALE)t1=MAX_COLOR_SCALE-1;
    color=t1;
    lir_setpixel(ix,line,color_scale[color]);
    i++;
    }
  }    
else
  {
  if(points_per_pixel == 0)
    {
// There are more pixels than data points so we must interpolate.
    i=fft2_first_point;
    yval=get_power(ptr+i);
    t1=log10(yval);
    t1=waterf_cfac*(t1-waterf_czer);
    if(t1<0)t1=0;
    if(t1 >= MAX_COLOR_SCALE)t1=MAX_COLOR_SCALE-1;
    color=t1;
    lir_setpixel(waterf_first_pixel,line,color_scale[color]);
    for(ix=waterf_first_pixel; ix < waterf_last_pixel; ix+=pixels_per_point)
      {
      i++;
      t2=get_power(ptr+i);
      der=(t2-yval)/pixels_per_point;
      for(k=ix+1; k<=ix+pixels_per_point; k++)
        {
        yval+=der;
        t1=log10(yval);
        t1=waterf_cfac*(t1-waterf_czer);
        if(t1<0)t1=0;
        if(t1 >= MAX_COLOR_SCALE)t1=MAX_COLOR_SCALE-1;
        color=t1;
        lir_setpixel(k,line,color_scale[color]);
        }
      yval=t2;
      }
    }  
  else
    {
// There is more than one data point for each pixel.
// Make a region ia to ib for the points behind each pixel.
// find the largest value for each region.
    ia=fft2_first_point;
    ib=ia+points_per_pixel;
    for(ix=waterf_first_pixel; ix < waterf_last_pixel; ix++)
      {
      t2=0;
      k=0;
      for(i=ia; i<ib; i++)
        {    
        t1=get_power(ptr+i);
        if(t1 > t2)t2=t1;
        }
      t1=log10(t2);
      t1=waterf_cfac*(t1-waterf_czer);
      if(t1<0)t1=0;
      if(t1 >= MAX_COLOR_SCALE)t1=MAX_COLOR_SCALE-1;
      color=t1;
      lir_setpixel(ix,line,color_scale[color]);
      ia=ib;
      ib+=points_per_pixel;
      if(ib >= fft2_size)ib=fft2_size;
      }
    }
  }
}

void scroll_waterfall(void)
{
int i,k;
// When arriving here the line (waterf_line-1) corresponds to the 
// latest power spectrum pointed to by (pwr_pa-fft2_size).
// redraw the lines at their new positions.
i=waterf_line+waterf_scroll;
k=pwr_pa;
if(!help_flag)lir_fillbox(waterf_first_pixel,
                               waterf_line+1,waterf_totpix,waterf_scroll,0);
waterf_line+=waterf_scroll;
while(i<waterf_last_line-1)
  {
  i++;
  draw_waterf_line(i,k);
  k-=fft2_size;
  if(k<0)k+=pwr_totsize;
  }
lir_refresh_screen();
}

void fresh_waterfall(void)
{
int i,k;
// This routine is called when parameters are changed in such a
// way that everything has to be recalculated.
lir_fillbox(0,waterf_first_line,screen_width-1,no_of_waterf_lines+1,0);
lir_hline(waterf_first_pixel-1,waterf_first_line,waterf_last_pixel,15);
lir_hline(waterf_first_pixel-1,waterf_last_line,waterf_last_pixel,15);
lir_line(waterf_first_pixel-1,waterf_first_line,
         waterf_first_pixel-1,waterf_last_line,15);
lir_line(waterf_last_pixel,waterf_first_line,
         waterf_last_pixel,waterf_last_line,15);
i=waterf_line;
k=pwr_pa;
while(i<waterf_last_line-1)
  {
  i++;
  k-=fft2_size;
  if(k<0)k+=pwr_totsize;
  draw_waterf_line(i,k);
  }
lir_refresh_screen();
}

void show_fq_scale(void)
{
int j;
double fq_valstep, dt1;
double fq_value, fq_x, fq_xstep;
// Express frequencies in Hz and do not place numbers closer than
// 6 ascii characters.
fq_valstep=8*waterf_hz_per_pixel*text_width;
j=adjust_scale(&fq_valstep);
// Get the decimal fraction and convert it to Hz  
dt1=zoom_freq*1000000.-zoom_width_points*waterf_hz_per_point;
dt1-=100000*(int)(dt1/100000.);
fq_value=fq_valstep*(int)(0.5+dt1/fq_valstep);
fq_x=waterf_first_pixel+(fq_value-dt1)/waterf_hz_per_pixel;
fq_xstep=fq_valstep/waterf_hz_per_pixel;
lir_fillbox(0,0,screen_width-1,waterf_first_line,0);
fq_scale(waterf_first_pixel, waterf_last_pixel-1, 0, j, 
                                       fq_valstep, fq_value, fq_x, fq_xstep);
}

void help(void)
{
char s[120];
clear_keyboard();
help_flag=TRUE;
clear_screen();
sprintf(s,"%s A program to wiev and zoom waterfall graphs",PROGRAM_NAME);
lir_text(0,0,s);
lir_text(0,1,"Move focus with the arrow keys. Set scales by use of the mouse");
lir_text(0,2,"or with the alpha-numeric keys as follows:");
lir_text(0,3,"    Scale   Increase  Decrease   Expand  Contract");
lir_text(0,4,"X=Frequency   '<-'      '->'       'Q'     'W'");
lir_text(0,5,"Y=Time        'up'     'dwn'       'E'     'R'");
lir_text(0,6,"Z=Colour      'A'       'S'        'D'     'F'");
lir_text(0,7,"Toggle polarization mode with the 'Z' key");
lir_refresh_screen();
while(lir_inkey == 0)
  {
  lir_sleep(5000);
  update_disk();
  if(kill_all_flag)return;
  test_keyboard();
  }
help_flag=FALSE;
show_fq_scale();
fresh_waterfall();
}

void set_zoom_endpoints(void)
{
int j;
zoom_width_pixels=screen_width/2-2;
if(pixels_per_point != 0)
  {
  waterf_hz_per_pixel=waterf_hz_per_point/pixels_per_point;
  zoom_width_points=zoom_width_pixels/pixels_per_point;
  if(zoom_width_points > fft2_size/2)zoom_width_points=fft2_size;
  zoom_width_pixels=zoom_width_points*pixels_per_point;
  }
else
  {
  waterf_hz_per_pixel=waterf_hz_per_point*points_per_pixel;
  zoom_width_points=zoom_width_pixels*points_per_pixel;
  if(zoom_width_points > fft2_size/2)zoom_width_points=fft2_size/2;
  zoom_width_pixels=zoom_width_points/points_per_pixel;
  zoom_width_points=zoom_width_pixels*points_per_pixel;
  }
fft2_first_point=fft2_size/2-zoom_width_points;
fft2_last_point=fft2_size/2+zoom_width_points;
j=1000000.*(zoom_freq-passband_center)/waterf_hz_per_point;
if(fft2_first_point+j <= 1)
  {
  zoom_freq=passband_center-waterf_hz_per_point*(fft2_first_point-1)/1000000.;
  }
if(fft2_last_point+j >= fft2_size-2)
  {
  zoom_freq=passband_center+waterf_hz_per_point*
                                    (fft2_size-2-fft2_last_point)/1000000.;
  }
j=1000000*(zoom_freq-passband_center)/waterf_hz_per_point;
fft2_first_point+=j;
fft2_last_point+=j;  
zoom_freq=passband_center+(waterf_hz_per_point*j)/1000000;

waterf_first_pixel=screen_width/2-zoom_width_pixels;
waterf_last_pixel=screen_width/2+zoom_width_pixels;
waterf_totpix=zoom_width_pixels*2;
show_fq_scale();
}

void show_scro_frame(int no)
{
lir_hline(scro[no].x1,scro[no].y1,scro[no].x2,10);
lir_hline(scro[no].x1,scro[no].y2,scro[no].x2,10);
lir_line(scro[no].x1,scro[no].y1,scro[no].x1,scro[no].y2,10);
lir_line(scro[no].x2,scro[no].y1,scro[no].x2,scro[no].y2,10);
}


void clear_write_scro(int no,char *s)
{
lir_fillbox(scro[no].x1+1, scro[no].y1+1,
                    scro[no].x2-scro[no].x1-2, scro[no].y2-scro[no].y1-2,0);
lir_pixwrite(scro[no].x1+1,scro[no].y1+2,s);
}

void fillwait(int no)
{
char s[80];
sprintf(s,"WAIT");
clear_write_scro(no,s);
}

void show_zoom_freq(void)
{
char s[80];
sprintf(s,"X0=%.6fMHz",zoom_freq);
clear_write_scro(ZOOM_FREQ,s);
}

void show_zoom_x_scale(void)
{
char s[80];
sprintf(s,"dX=%3.4fHz",waterf_hz_per_pixel);
clear_write_scro(ZOOM_HZ_PER_PIXEL,s);
}

void show_waterf_avgnum(void)
{
char s[80];
sprintf(s,"Avg %d",waterf_avgnum);
clear_write_scro(ZOOM_AVGNUM,s);
}

void show_colour_zero(void)
{
char s[80];
sprintf(s,"Color ref %.2f",10*log10(waterf_czer));
clear_write_scro(ZOOM_COLOUR_ZERO,s);
}

void show_colour_fac(void)
{
char s[80];
sprintf(s,"Color gain %.2f",waterf_cfac);
clear_write_scro(ZOOM_COLOUR_FAC,s);
}




void show_delay_time(void)
{
int tim,min,sec;
char s[80];
if(real_time_flag)
  {
  sec=fft2_block_time*((block_pa-block_pw+block_totsize)&block_mask);
  }
else
  {
  sec=0;
  }  
sec=fft2_block_time*((block_pa-block_pw+block_totsize)&block_mask);
tim=sec/3600;
sec-=3600*tim;
min=sec/60;
sec-=60*min;
sprintf(s,"Del %d.%d.%d",tim,min,sec);
clear_write_scro(ZOOM_DELAY_TIME,s);
}

void show_pol_mode(void)
{
if(no_of_rx_channels==1)return;
clear_write_scro(ZOOM_POL_MODE, pol_mode_text[pol_mode]);
}


void make_scro(void)
{
scro[ZOOM_FREQ].x1=0;
scro[ZOOM_FREQ].y1=waterf_last_line+2;
scro[ZOOM_FREQ].x2=scro[ZOOM_FREQ].x1+22*text_width+2;
scro[ZOOM_FREQ].y2=scro[ZOOM_FREQ].y1+text_height;
show_scro_frame(ZOOM_FREQ);
show_zoom_freq();
scro[ZOOM_HZ_PER_PIXEL].x1=scro[ZOOM_FREQ].x2+2*text_width;
scro[ZOOM_HZ_PER_PIXEL].y1=waterf_last_line+2;
scro[ZOOM_HZ_PER_PIXEL].x2=scro[ZOOM_HZ_PER_PIXEL].x1+16*text_width+2;
scro[ZOOM_HZ_PER_PIXEL].y2=scro[ZOOM_HZ_PER_PIXEL].y1+text_height;
show_scro_frame(ZOOM_HZ_PER_PIXEL);
show_zoom_x_scale();
scro[ZOOM_AVGNUM].x1=scro[ZOOM_HZ_PER_PIXEL].x2+2*text_width;
scro[ZOOM_AVGNUM].y1=waterf_last_line+2;
scro[ZOOM_AVGNUM].x2=scro[ZOOM_AVGNUM].x1+11*text_width+2;
scro[ZOOM_AVGNUM].y2=scro[ZOOM_AVGNUM].y1+text_height;
show_scro_frame(ZOOM_AVGNUM);
show_waterf_avgnum();
scro[ZOOM_COLOUR_ZERO].x1=0;
scro[ZOOM_COLOUR_ZERO].y1=text_height+waterf_last_line+4;
scro[ZOOM_COLOUR_ZERO].x2=scro[ZOOM_COLOUR_ZERO].x1+22*text_width+2;
scro[ZOOM_COLOUR_ZERO].y2=scro[ZOOM_COLOUR_ZERO].y1+text_height;
show_scro_frame(ZOOM_COLOUR_ZERO);
show_colour_zero();
scro[ZOOM_COLOUR_FAC].x1=scro[ZOOM_COLOUR_ZERO].x2+text_width+2;
scro[ZOOM_COLOUR_FAC].y1=text_height+waterf_last_line+4;
scro[ZOOM_COLOUR_FAC].x2=scro[ZOOM_COLOUR_FAC].x1+22*text_width+2;
scro[ZOOM_COLOUR_FAC].y2=scro[ZOOM_COLOUR_FAC].y1+text_height;
show_scro_frame(ZOOM_COLOUR_FAC);
show_colour_fac();
scro[ZOOM_DELAY_TIME].x1=scro[ZOOM_COLOUR_FAC].x2+text_width+2;
scro[ZOOM_DELAY_TIME].y1=text_height+waterf_last_line+4;
scro[ZOOM_DELAY_TIME].x2=scro[ZOOM_DELAY_TIME].x1+22*text_width+2;
scro[ZOOM_DELAY_TIME].y2=scro[ZOOM_DELAY_TIME].y1+text_height;
show_scro_frame(ZOOM_DELAY_TIME);
show_delay_time();
scro[ZOOM_POL_MODE].x1=0;
scro[ZOOM_POL_MODE].y1=2*text_height+waterf_last_line+4;
scro[ZOOM_POL_MODE].x2=scro[ZOOM_POL_MODE].x1+15*text_width+2;
scro[ZOOM_POL_MODE].y2=scro[ZOOM_POL_MODE].y1+text_height;
show_scro_frame(ZOOM_POL_MODE);
show_pol_mode();
no_of_scro=ZOOM_MAX_SCRO;
}

void increase_zoom_freq(void)
{
fillwait(ZOOM_FREQ);
zoom_freq+=(0.5*zoom_width_points*waterf_hz_per_point)/1000000;
set_zoom_endpoints();
show_zoom_freq();
fresh_waterfall();
}

void decrease_zoom_freq(void)
{
fillwait(ZOOM_FREQ);
zoom_freq-=(0.5*zoom_width_points*waterf_hz_per_point)/1000000;
set_zoom_endpoints();
show_zoom_freq();
fresh_waterfall();
}



void update_disk(void)
{
int i;
int fno, pos;
float t1,t2,t3,t4,t5,t6;
float lin0, total;
while(fft2_px != fft2_pa)
  {
  fno=block_pa/blocks_per_file;
  pos=((block_pa&file_mask)*block_bytes);
  fseek(tmpfile_fd[fno],pos,SEEK_SET);
  if(swfloat)
    {
    fwrite(&fft2_float[fft2_px],fft2_block,sizeof(float),
                                                   tmpfile_fd[fno]);
    }
  else
    {
    fwrite(&fft2_shi[fft2_px],fft2_block,sizeof(short int),
                                                   tmpfile_fd[fno]);
    }
  if(no_of_rx_channels==2)
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        lin0=fft2_float[fft2_px+4*i  ]*fft2_float[fft2_px+4*i  ]+
             fft2_float[fft2_px+4*i+1]*fft2_float[fft2_px+4*i+1];
        total=fft2_float[fft2_px+4*i+2]*fft2_float[fft2_px+4*i+2]+
              fft2_float[fft2_px+4*i+3]*fft2_float[fft2_px+4*i+3]+lin0;
        t1=fft2_float[fft2_px+4*i  ]+fft2_float[fft2_px+4*i+3];
        t2=fft2_float[fft2_px+4*i+1]-fft2_float[fft2_px+4*i+2];
        t3=COS30*fft2_float[fft2_px+4*i  ]+SIN30*fft2_float[fft2_px+4*i+3];
        t4=COS30*fft2_float[fft2_px+4*i+1]+SIN30*fft2_float[fft2_px+4*i+2];
        t5=COS60*fft2_float[fft2_px+4*i  ]+SIN60*fft2_float[fft2_px+4*i+3];
        t6=COS60*fft2_float[fft2_px+4*i+1]+SIN60*fft2_float[fft2_px+4*i+2];
        }
      else
        {
        lin0=(float)(fft2_shi[fft2_px+4*i  ])*fft2_shi[fft2_px+4*i  ]+
             (float)(fft2_shi[fft2_px+4*i+1])*fft2_shi[fft2_px+4*i+1];
        total=(float)(fft2_shi[fft2_px+4*i+2])*fft2_shi[fft2_px+4*i+2]+
              (float)(fft2_shi[fft2_px+4*i+3])*fft2_shi[fft2_px+4*i+3]+lin0;
        t1=(float)(fft2_shi[fft2_px+4*i  ])+(float)(fft2_shi[fft2_px+4*i+3]);
        t2=(float)(fft2_shi[fft2_px+4*i+1])-(float)(fft2_shi[fft2_px+4*i+2]);
        t3=COS30*fft2_shi[fft2_px+4*i  ]+SIN30*fft2_shi[fft2_px+4*i+3];
        t4=COS30*fft2_shi[fft2_px+4*i+1]+SIN30*fft2_shi[fft2_px+4*i+2];
        t5=COS60*fft2_shi[fft2_px+4*i  ]+SIN60*fft2_shi[fft2_px+4*i+3];
        t6=COS60*fft2_shi[fft2_px+4*i+1]+SIN60*fft2_shi[fft2_px+4*i+2];
        }
      if(block_pb == block_pa)
        {
        fft2_xysum[pwr_pa+i].lin0=lin0;
        fft2_xysum[pwr_pa+i].total=total;
        fft2_xysum[pwr_pa+i].circ=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_pa+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_pa+i].lin60=t5*t5+t6*t6; 
        }
      else
        {
        fft2_xysum[pwr_pa+i].lin0+=lin0;
        fft2_xysum[pwr_pa+i].total+=total;
        fft2_xysum[pwr_pa+i].circ+=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_pa+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_pa+i].lin60=t5*t5+t6*t6; 
        }
      }                                             
    }
  else    
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        t1=fft2_float[fft2_px+2*i  ];
        t2=fft2_float[fft2_px+2*i+1];
        }
      else
        {
        t1=fft2_shi[fft2_px+2*i  ];
        t2=fft2_shi[fft2_px+2*i+1];
        }
      if(block_pb == block_pa)
        {
        fft2_powersum[pwr_pa+i]=t1*t1+t2*t2;
        }
      else
        {
        fft2_powersum[pwr_pa+i]+=t1*t1+t2*t2;
        }
      }    
    }
  block_pa=(block_pa+1)&block_mask;
  if(block_px == block_pa)block_px=(block_px+1)&block_mask;
  if(real_time_flag)
    {
    if( ((block_pa-block_pb+block_totsize)&block_mask) >= waterf_avgnum)
      {
      draw_waterf_line(waterf_line,pwr_pa);
      waterf_line--;
      if(waterf_line <= waterf_first_line)
        {
        scroll_waterfall();
        }
      pwr_pa+=fft2_size;
      if(pwr_pa >= pwr_totsize)pwr_pa=0;
      if(pwr_px==pwr_pa)
        {
        pwr_px+=fft2_size;
        if(pwr_px >= pwr_totsize)pwr_px=0;
        }
      block_pb=block_pa;
      }
    }
  fft2_px=(fft2_px+fft2_block)&fft2_mask;  
  }
lir_refresh_screen();
}

void await_completed_line(void)
{
if(real_time_flag)
  {
  lir_text(0,screen_last_line,"Awaiting completed average.");
  while(block_pb != block_pa)
    {
    lir_sleep(5000);
    update_disk();
    if(kill_all_flag)return;
    }
  clear_lines(screen_last_line,screen_last_line);  
  }  
}

void recompute_old_powers(void)
{
int i, k, avgn;
int fno, pos;
int blk_ptr;
int no;
float t1,t2,t3,t4,t5,t6;
float lin0, total;
lir_text(0,screen_last_line,"Awaiting new powers computed from disk.");
// Powers from pwr_pa-fft2_size to pwr_px are already correct.
// Use the data from scratch files to compute older data.
no=pwr_pa-pwr_px;
if(no < 0)no+=pwr_totsize;
no/=fft2_size;
no++;
blk_ptr=(block_pw-(no-1)*waterf_avgnum+block_totsize)&block_mask;
avgn=0;
while(pwr_px != pwr_pa)
  {
  fno=blk_ptr/blocks_per_file;
  pos=((blk_ptr&file_mask)*block_bytes);
  fseek(tmpfile_fd[fno],pos,SEEK_SET);
  if(swfloat)
    {
    fread(tmp_float,fft2_block,sizeof(float),tmpfile_fd[fno]);
    }
  else
    {
    fread(tmp_shi,fft2_block,sizeof(short int),tmpfile_fd[fno]);
    }
  if(no_of_rx_channels==2)
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        lin0=tmp_float[4*i  ]*tmp_float[4*i  ]+
             tmp_float[4*i+1]*tmp_float[4*i+1];
        total=tmp_float[4*i+2]*tmp_float[4*i+2]+
              tmp_float[4*i+3]*tmp_float[4*i+3]+lin0;
        t1=tmp_float[4*i  ]+tmp_float[4*i+3];
        t2=tmp_float[4*i+1]-tmp_float[4*i+2];
        t3=COS30*tmp_float[4*i  ]+SIN30*tmp_float[4*i+3];
        t4=COS30*tmp_float[4*i+1]+SIN30*tmp_float[4*i+2];
        t5=COS60*tmp_float[4*i  ]+SIN60*tmp_float[4*i+3];
        t6=COS60*tmp_float[4*i+1]+SIN60*tmp_float[4*i+2];
        }
      else
        {
        lin0=(float)(tmp_shi[4*i  ])*tmp_shi[4*i  ]+
             (float)(tmp_shi[4*i+1])*tmp_shi[4*i+1];
        total=(float)(tmp_shi[4*i+2])*tmp_shi[4*i+2]+
              (float)(tmp_shi[4*i+3])*tmp_shi[4*i+3]+lin0;
        t1=(float)(tmp_shi[4*i  ])+(float)(tmp_shi[4*i+3]);
        t2=(float)(tmp_shi[4*i+1])-(float)(tmp_shi[4*i+2]);
        t3=COS30*tmp_shi[4*i  ]+SIN30*tmp_shi[4*i+3];
        t4=COS30*tmp_shi[4*i+1]+SIN30*tmp_shi[4*i+2];
        t5=COS60*tmp_shi[4*i  ]+SIN60*tmp_shi[4*i+3];
        t6=COS60*tmp_shi[4*i+1]+SIN60*tmp_shi[4*i+2];
        }
      if(avgn == 0)
        {
        fft2_xysum[pwr_px+i].lin0=lin0;
        fft2_xysum[pwr_px+i].total=total;
        fft2_xysum[pwr_px+i].circ=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_px+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_px+i].lin60=t5*t5+t6*t6; 
        }
      else
        {
        fft2_xysum[pwr_px+i].lin0+=lin0;
        fft2_xysum[pwr_px+i].total+=total;
        fft2_xysum[pwr_px+i].circ+=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_px+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_px+i].lin60=t5*t5+t6*t6; 
        }
      }                                             
    }
  else    
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        t1=tmp_float[2*i  ];
        t2=tmp_float[2*i+1];
        }
      else
        {
        t1=tmp_shi[2*i  ];
        t2=tmp_shi[2*i+1];
        }
      if(avgn == 0)
        {
        fft2_powersum[pwr_px+i]=t1*t1+t2*t2;
        }
      else
        {
        fft2_powersum[pwr_px+i]+=t1*t1+t2*t2;
        }
      }    
    }
  avgn++;  
  blk_ptr=(blk_ptr-1)&block_mask;
  if(blk_ptr==block_px)goto recompute_old_x;    
  if(avgn >= waterf_avgnum)
    {
    avgn=0;
    pwr_px-=fft2_size;
    if(pwr_px < 0)pwr_px+=pwr_totsize;
    }
  k=pwr_px;
  if( ((fft2_pa-fft2_px+fft2_totsize)&fft2_mask) >= (fft2_totsize>>1))
    {
    update_disk();
    }
  if(k!=pwr_px || blk_ptr==block_px)goto recompute_old_x;
  if(kill_all_flag)return;
  }   
recompute_old_x:;   
if(pwr_px == pwr_pa)
  {
  pwr_px+=fft2_size;
  if(pwr_px >= pwr_totsize)pwr_px=0;
  }
clear_lines(screen_last_line,screen_last_line);
}

void recompute_new_powers(void)
{
int i, avgn;
int fno, pos;
int blk_ptr, pwr_ptr;
int no;
float t1,t2,t3,t4,t5,t6;
float lin0, total;
// Powers from pwr_pa-fft2_size to pwr_pp are not correct.
// Use the data from scratch files to compute data.
if(real_time_flag)
  {
  blk_ptr=(block_pb-1+block_totsize)&block_mask;
  }
else
  {  
  no=pwr_pa-pwr_pp;
  if(no < 0)no+=pwr_totsize;
  no/=fft2_size;
  blk_ptr=(block_pw-(no-1)*waterf_avgnum+block_totsize)&block_mask;
  }
avgn=0;
pwr_ptr=pwr_pp;
while(pwr_ptr != pwr_pa)
  {
  fno=blk_ptr/blocks_per_file;
  pos=((blk_ptr&file_mask)*block_bytes);
  fseek(tmpfile_fd[fno],pos,SEEK_SET);
  if(swfloat)
    {
    fread(tmp_float,fft2_block,sizeof(float),tmpfile_fd[fno]);
    }
  else
    {
    fread(tmp_shi,fft2_block,sizeof(short int),tmpfile_fd[fno]);
    }
  if(no_of_rx_channels==2)
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        lin0=tmp_float[4*i  ]*tmp_float[4*i  ]+
             tmp_float[4*i+1]*tmp_float[4*i+1];
        total=tmp_float[4*i+2]*tmp_float[4*i+2]+
              tmp_float[4*i+3]*tmp_float[4*i+3]+lin0;
        t1=tmp_float[4*i  ]+tmp_float[4*i+3];
        t2=tmp_float[4*i+1]-tmp_float[4*i+2];
        t3=COS30*tmp_float[4*i  ]+SIN30*tmp_float[4*i+3];
        t4=COS30*tmp_float[4*i+1]+SIN30*tmp_float[4*i+2];
        t5=COS60*tmp_float[4*i  ]+SIN60*tmp_float[4*i+3];
        t6=COS60*tmp_float[4*i+1]+SIN60*tmp_float[4*i+2];
        }
      else
        {
        lin0=(float)(tmp_shi[4*i  ])*tmp_shi[4*i  ]+
             (float)(tmp_shi[4*i+1])*tmp_shi[4*i+1];
        total=(float)(tmp_shi[4*i+2])*tmp_shi[4*i+2]+
              (float)(tmp_shi[4*i+3])*tmp_shi[4*i+3]+lin0;
        t1=(float)(tmp_shi[4*i  ])+(float)(tmp_shi[4*i+3]);
        t2=(float)(tmp_shi[4*i+1])-(float)(tmp_shi[4*i+2]);
        t3=COS30*tmp_shi[4*i  ]+SIN30*tmp_shi[4*i+3];
        t4=COS30*tmp_shi[4*i+1]+SIN30*tmp_shi[4*i+2];
        t5=COS60*tmp_shi[4*i  ]+SIN60*tmp_shi[4*i+3];
        t6=COS60*tmp_shi[4*i+1]+SIN60*tmp_shi[4*i+2];
        }
      if(avgn == 0)
        {
        fft2_xysum[pwr_ptr+i].lin0=lin0;
        fft2_xysum[pwr_ptr+i].total=total;
        fft2_xysum[pwr_ptr+i].circ=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_ptr+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_ptr+i].lin60=t5*t5+t6*t6; 
        }
      else
        {
        fft2_xysum[pwr_ptr+i].lin0+=lin0;
        fft2_xysum[pwr_ptr+i].total+=total;
        fft2_xysum[pwr_ptr+i].circ+=0.5*(t1*t1+t2*t2); 
        fft2_xysum[pwr_ptr+i].lin30=t3*t3+t4*t4; 
        fft2_xysum[pwr_ptr+i].lin60=t5*t5+t6*t6; 
        }
      }                                             
    }
  else    
    {
    for(i=0; i<fft2_size; i++)
      {
      if(swfloat)
        {
        t1=tmp_float[2*i  ];
        t2=tmp_float[2*i+1];
        }
      else
        {
        t1=tmp_shi[2*i  ];
        t2=tmp_shi[2*i+1];
        }
      if(avgn == 0)
        {
        fft2_powersum[pwr_ptr+i]=t1*t1+t2*t2;
        }
      else
        {
        fft2_powersum[pwr_ptr+i]+=t1*t1+t2*t2;
        }
      }    
    }
  avgn++;  
  blk_ptr=(blk_ptr+1)&block_mask;
  if(avgn >= waterf_avgnum)
    {
    avgn=0;
    pwr_ptr+=fft2_size;
    if(pwr_ptr >= pwr_totsize)pwr_ptr-=pwr_totsize;
    }
  if( ((fft2_pa-fft2_px+fft2_totsize)&fft2_mask) >= (fft2_totsize>>1))
    {
    update_disk();
    }
  if(kill_all_flag)return;
  }   
}

void decrease_time_delay(void)
{
int m, no, shift, k;
if(real_time_flag)return;
shift=no_of_waterf_lines>>2;
fillwait(ZOOM_DELAY_TIME);
pwr_pp=pwr_pa;
dectim_w:;
no=(block_pa-block_pw+block_totsize)&block_mask;
if(no <= waterf_avgnum*shift)
  {
  k=no/waterf_avgnum;
  if(k*waterf_avgnum != no)
    {
    lir_sleep(1000);
    update_disk();
    goto dectim_w;
    }
  block_pb=block_pa;
  real_time_flag=TRUE;
  waterf_line=waterf_first_line+waterf_scroll;
  }
else
  {
  block_pw=(block_pw+waterf_avgnum*shift)&block_mask;
  k=shift;
  }
m=pwr_px-pwr_pa;
if(m < 0)m+=pwr_totsize;
m/=fft2_size;
pwr_pa+=k*fft2_size;
if(pwr_pa >= pwr_totsize)pwr_pa-=pwr_totsize;
if(m<=k)
  {
  pwr_px=pwr_pa+fft2_size;
  if(pwr_px >= pwr_totsize)pwr_px-=pwr_totsize;
  }  
waterf_line=waterf_first_line+1;
fresh_waterfall();
recompute_new_powers();
show_delay_time();
fresh_waterfall();
}




void increase_time_delay(void)
{
int no, pp;
int shift;
no=pwr_pa-pwr_px;
if(no < 0)no+=pwr_totsize;
no/=fft2_size;
shift=no_of_waterf_lines>>2;
if(no < 5*(shift+1))return;
fillwait(ZOOM_DELAY_TIME);
if(real_time_flag)
  {
  real_time_flag=FALSE;
  pp=block_pb;
  }
else
  {
  pp=block_pw;
  }
block_pw=(pp-waterf_avgnum*shift+block_totsize)&block_mask;
waterf_line=waterf_first_line+1;
pwr_pa-=shift*fft2_size;
if(pwr_pa < 0)pwr_pa+=pwr_totsize;
fresh_waterfall();
recompute_old_powers();
show_delay_time();
fresh_waterfall();
}

void finish_avgnum_change(void)
{
fresh_waterfall();
if(real_time_flag)
  {
  await_completed_line();
  block_pw=block_pb;
  }
pwr_px=pwr_pa-fft2_size;
if(pwr_px < 0)pwr_px+=pwr_totsize;
recompute_old_powers();
show_waterf_avgnum();
fresh_waterfall();
}

void reduce_waterf_avgnum(void)
{
int no;
int i, pa, pb, pc;
if(waterf_avgnum == 1)return;
fillwait(ZOOM_AVGNUM);
if(real_time_flag)await_completed_line();
// The latest line is at pwr_pa-fft2_size.
// Split the power into two equal parts and put into
// the positions pwr_pa-fft2_size and pwr_pa-2*fft2_size
// The current line is waterf_line and we nead to treat
// all lines down to waterf_last_line.
no=(1+waterf_last_line-waterf_line)/2;
if(waterf_line-waterf_first_line>waterf_scroll)
  {
  waterf_line-=2*no-1;
  if(waterf_line<waterf_scroll)
    {
    waterf_line=waterf_scroll+waterf_first_line;
    }
  no=(1+waterf_last_line-waterf_line)/2;
  }  
pa=pwr_pa-no*fft2_size;   
if(pa < 0)pa+=pwr_totsize;
pb=pwr_pa-2*(no)*fft2_size;     
if(pb < 0)pb+=pwr_totsize;
while(pa != pwr_pa)
  {
  pc=pb+fft2_size;
  if(pc>pwr_totsize)pc-=pwr_totsize;
  if(no_of_rx_channels==2)
    {
    for(i=0; i<fft2_size; i++)
      {
      fft2_xysum[pb+i].lin0=fft2_xysum[pa+i].lin0*0.5;
      fft2_xysum[pb+i].lin30=fft2_xysum[pa+i].lin30*0.5;
      fft2_xysum[pb+i].lin60=fft2_xysum[pa+i].lin60*0.5;
      fft2_xysum[pb+i].circ=fft2_xysum[pa+i].circ*0.5;
      fft2_xysum[pb+i].total=fft2_xysum[pa+i].total*0.5;
      fft2_xysum[pc+i]=fft2_xysum[pb+i];
      }
    }
  else
    {
    for(i=0; i<fft2_size; i++)
      {
      fft2_powersum[pb+i]=fft2_powersum[pa+i]*0.5;
      fft2_powersum[pc+i]=fft2_powersum[pb+i];
      }
    }
  pb=pb+2*fft2_size;
  if(pb >= pwr_totsize)pb-=pwr_totsize;
  pa=pa+fft2_size;
  if(pa >= pwr_totsize)pa=0;
  }
waterf_avgnum/=2;
finish_avgnum_change();
}

void increase_waterf_avgnum(void)
{
int no;
int i, pa, pb, pc;
fillwait(ZOOM_AVGNUM);
if(real_time_flag)await_completed_line();
// The latest line is at pwr_pa-fft2_size.
// Sum lines pair-wise to contract the waterfall.
// The oldest line with valid data is at pwr_px.
no=pwr_pa-pwr_px;
if(no < 0)no+=pwr_totsize;
no/=2*fft2_size;
if(no<waterf_last_line-waterf_line)waterf_line=waterf_last_line-no;
pwr_px=pwr_pa-no*fft2_size;
if(pwr_px < 0)pwr_px+=pwr_totsize;
pa=pwr_pa-fft2_size;   
if(pa < 0)pa+=pwr_totsize;
pb=pa;
while(pa != pwr_px)
  {
  pc=pb-fft2_size;
  if(pc<0)pc+=pwr_totsize;
  if(no_of_rx_channels==2)
    {
    for(i=0; i<fft2_size; i++)
      {
      fft2_xysum[pa+i].lin0=fft2_xysum[pb+i].lin0+fft2_xysum[pc+i].lin0;
      fft2_xysum[pa+i].lin30=fft2_xysum[pb+i].lin30+fft2_xysum[pc+i].lin30;
      fft2_xysum[pa+i].lin60=fft2_xysum[pb+i].lin60+fft2_xysum[pc+i].lin60;
      fft2_xysum[pa+i].circ=fft2_xysum[pb+i].circ+fft2_xysum[pc+i].circ;
      fft2_xysum[pa+i].total=fft2_xysum[pb+i].total+fft2_xysum[pc+i].total;
      }
    }
  else
    {
    for(i=0; i<fft2_size; i++)
      {
      fft2_powersum[pa+i]=fft2_powersum[pb+i]+fft2_powersum[pc+i];
      }
    }
  pb=pb-2*fft2_size;
  if(pb<0)pb+=pwr_totsize;
  pa=pa-fft2_size;
  if(pa<0)pa+=pwr_totsize;
  no--;
  }
waterf_avgnum*=2;
finish_avgnum_change();
}

void increase_pixels_per_point(void)
{
fillwait(ZOOM_HZ_PER_PIXEL);
lir_refresh_screen();
if(pixels_per_point > 0)
  {
  pixels_per_point*=2;
  points_per_pixel=0;
  }
else
  {
  points_per_pixel/=2;
  if(points_per_pixel == 1)pixels_per_point=1;
  }
set_zoom_endpoints();
show_zoom_x_scale();
fresh_waterfall();
}

void decrease_pixels_per_point(void)
{
fillwait(ZOOM_HZ_PER_PIXEL);
lir_refresh_screen();
if(pixels_per_point > 1)
  {
  pixels_per_point/=2;
  if(pixels_per_point==1)points_per_pixel=1;
  }
else
  {
  points_per_pixel*=2;
  pixels_per_point=0;
  }
set_zoom_endpoints();
show_zoom_x_scale();
fresh_waterfall();
}


void reduce_waterf_czer(void)
{
waterf_czer-=0.25*MAX_COLOR_SCALE/waterf_cfac;
show_colour_zero();
fresh_waterfall();
}

void increase_waterf_czer(void)
{
waterf_czer+=0.25*MAX_COLOR_SCALE/waterf_cfac;
show_colour_zero();
fresh_waterfall();
}

void increase_waterf_cfac(void)
{
waterf_cfac*=1.25;
show_colour_fac();
fresh_waterfall();
}

void decrease_waterf_cfac(void)
{
waterf_cfac*=0.75;
show_colour_fac();
fresh_waterfall();
}

void toggle_pol_mode(void)
{
if(no_of_rx_channels==1)return;
pol_mode++;
if(pol_mode >= POL_MAX_TYPES)pol_mode=0;
show_pol_mode();
fresh_waterfall();
}


void zoom(void)
{
int i;
lir_refresh_screen();
fft2_float=NULL;
fft2_shi=NULL;
tmp_float=NULL;
tmp_shi=NULL;
fft2_powersum=NULL;
fft2_xysum=NULL;
waterf_last_line=screen_height-4*(text_height+4);
waterf_first_line=(text_height+5)&-2;
no_of_waterf_lines=waterf_last_line-waterf_first_line;
if( no_of_waterf_lines < 50)
  {
  fatal_error("Screen too small");
  goto zoom_x;
  }
if(no_of_waterf_lines > ui.max_lines)
  {
  no_of_waterf_lines=ui.max_lines;
  }
waterf_scroll=no_of_waterf_lines/16;
no_of_waterf_lines=16*waterf_scroll;
waterf_last_line=waterf_first_line+no_of_waterf_lines;
if(open_network()==FALSE)goto zoom_x;
zoom_freq=passband_center;
if(fft2_sinpow != 0)
  {
  fft2_interleave_ratio=2*asin(pow(0.5,1.0/fft2_sinpow))/ 3.14159265;
  }
else fft2_interleave_ratio=0;
fft2_size=1<<fft2_n;
twice_rxchan=2*no_of_rx_channels;
fft2_block=twice_rxchan*fft2_size;
// Make the buffer big enough to hold 5 seconds of data.
fft2_block_time=(1.-fft2_interleave_ratio)*fft2_size/timf2_sampling_speed;
fft2_totsize=timf2_sampling_speed*5/(1-fft2_interleave_ratio);
make_power_of_two(&fft2_totsize);
max_fft2n=fft2_totsize/fft2_block;
DEB"\nmax_fft2n=%d",max_fft2n);
fft2_mask=fft2_totsize-1;
fft2_pa=0;
fft2_px=0;
if(swfloat)
  {
  fft2_float=malloc((fft2_totsize+fft2_block)*sizeof(float));
  if(fft2_float==NULL)
    {
    fatal_error(out_of_memory);
    goto zoom_x;
    }
  tmp_float=&fft2_float[fft2_totsize];  
  DEB"\n(float)fft2 memory = %f megabytes",
                                     0.000001*fft2_totsize*sizeof(float));
  for(i=0; i<fft2_block; i++)fft2_float[i]=0;
  }  
else
  {
  fft2_shi=malloc((fft2_totsize+fft2_block)*sizeof(short int));
  if(fft2_shi==NULL)
    {
    fatal_error(out_of_memory);
    goto zoom_x;
    }
  for(i=0; i<fft2_block; i++)fft2_shi[i]=0;
  tmp_shi=&fft2_shi[fft2_totsize];  
  DEB"\n(short int)fft2 memory = %f megabytes",
                                 0.000001*fft2_totsize*sizeof(short int));
  }  
mouse_task_no=-1;
no_of_scro=0;
leftpressed=0;
rightpressed=0;
mouse_flag=1;
show_mouse();
for(i=0; i<NO_OF_TMPFILES; i++)
  {
  tmpfile_fd[i]=fopen(tmpfile_name[i],"w+b");
  if(tmpfile_fd[i]==NULL)
    {
    while(i > 0)
      {
      i--;
      fclose(tmpfile_fd[i]);
      }
    fatal_error("Could not open scratch file(s)");
    goto zoom_x;
    }
  }

file_size=1<<(ui.max_filesiz-4);
DEB"\nScratch file size=%f megabytes per file. (%d files).",
                                 0.000001*file_size,NO_OF_TMPFILES);  
blocks_per_file=file_size/fft2_block;
if(swfloat)
  {
  block_bytes=fft2_block*sizeof(float);
  blocks_per_file/=sizeof(float);
  }
else
  {
  block_bytes=fft2_block*sizeof(short int);
  blocks_per_file/=sizeof(short int);
  }
DEB"\nTransforms per file=%d",blocks_per_file); 
block_pa=0;
block_px=0;
block_pw=0;
pwr_pa=0;
pwr_px=0;
block_totsize=blocks_per_file*NO_OF_TMPFILES;
block_mask=block_totsize-1;
file_mask=blocks_per_file-1;
waterf_hz_per_point=(float)(timf2_sampling_speed)/fft2_size;
pixels_per_point=screen_width/fft2_size;
points_per_pixel=fft2_size/screen_width;
set_zoom_endpoints();
waterf_avgnum=2;
waterf_czer=1.0+log10((float)(waterf_avgnum));
if(swfloat)waterf_czer+=4.5;
pwr_totsize=2*no_of_waterf_lines+1;
pwr_totsize*=fft2_size;
waterf_line=waterf_last_line-1;
if(no_of_rx_channels == 1)
  {
  fft2_powersum=malloc(pwr_totsize*sizeof(float));
  if(fft2_powersum == NULL)fatal_error(out_of_memory);
  DEB"\npowers memory = %f megabytes",0.000001*pwr_totsize*sizeof(float));
  }
else
  {
  fft2_xysum=malloc(pwr_totsize*sizeof(TWOCHAN_POWER));
  if(fft2_xysum == NULL)fatal_error(out_of_memory);
  DEB"\nxypowers memory = %f megabytes",
                                 0.000001*pwr_totsize*sizeof(TWOCHAN_POWER));

  }  

help_flag=FALSE;
real_time_flag=TRUE;
make_scro();
fresh_waterfall();
create_netrec_fft2_thread();
while(!kill_all_flag)
  {
  lir_sleep(1000);
  update_disk();

  test_keyboard();
  if(lir_inkey != 0)
    {
    switch (toupper(lir_inkey))
      {
      case F1_KEY:
      help();
      make_scro();
      break;

      case ARROW_LEFT_KEY:
      increase_zoom_freq();
      break;

      case ARROW_RIGHT_KEY:
      decrease_zoom_freq();
      break;

      case ARROW_UP_KEY:
      increase_time_delay();
      break;

      case ARROW_DOWN_KEY:
      decrease_time_delay();
      break;

      case 'W':
      decrease_pixels_per_point();
      break;
      
      case 'Q':
      increase_pixels_per_point();
      break;
      
      case 'D':
      increase_waterf_cfac();
      break;
      
      case 'F':
      decrease_waterf_cfac();
      break;
      
      case 'A':
      reduce_waterf_czer();
      break;
      
      case 'S':
      increase_waterf_czer();
      break;
      
      case 'E':
      reduce_waterf_avgnum();
      break;

      case 'Z':
      toggle_pol_mode();
      break;      

      case 'R':
      increase_waterf_avgnum();
      break;
      
      }
    }
  if( new_mouse_x!=mouse_x || new_mouse_y!=mouse_y)
    {
    lir_move_mouse_cursor();
    show_mouse();
    }
  if(passband_center != old_passband_center)
    {
    old_passband_center=passband_center;
    set_zoom_endpoints();
    show_zoom_freq();
    }
  if(mouse_task_no != -1)
    {
    set_button_states();
    switch (mouse_task_no)
      {
      case ZOOM_FREQ:
      break;
        
      case ZOOM_HZ_PER_PIXEL:
      if( (mouse_task_info&GRAPH_RIGHTPRESSED) != 0)
        {
        increase_pixels_per_point();
        }
      else
        {
        decrease_pixels_per_point();
        }
      break;
        
      case ZOOM_AVGNUM:
      break;

      }  
      mouse_task_no=-1;
      while(lbutton_state !=0 || rbutton_state != 0)
        {
        set_button_states();
        lir_sleep(10000);
        update_disk();
        }
/*
    if(numinput_flag==0)
      {
      current_mouse_activity();
      if(kill_all_flag)return;
      }  
    if( (numinput_flag&DATA_READY_PARM) != 0)
      {
      par_from_keyboard_routine();
      if(kill_all_flag)return;
      par_from_keyboard_routine=NULL;
      mouse_active_flag=0;
      numinput_flag=0;
      leftpressed=BUTTON_IDLE;  
      }
    if(mouse_active_flag == 0)
      {
      mouse_task_no=-1;
      }
mouse_x:;
 */
    }
  }
mouse_flag=0;  
wait_for_netrec_exit();
for(i=0; i<NO_OF_TMPFILES; i++)
  {
  fclose(tmpfile_fd[i]);
  }
zoom_x:;  
for(i=0; i<NO_OF_TMPFILES; i++)
  {
  unlink(tmpfile_name[i]);
  }
chk_free(fft2_float);
chk_free(fft2_shi);
chk_free(fft2_powersum);
chk_free(fft2_xysum);
}
